"""
data_selector.py

A node and widget for subselecting from a dataset.
"""
from typing import List, Tuple, Dict, Any, Sequence, Optional

import numpy as np

from .node import Node, NodeWidget, updateOption
from ..data.datadict import DataDictBase, DataDict
from ..gui.data_display import DataSelectionWidget
from plottr.icons import get_dataColumnsIcon
from ..utils import num

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'


class DataDisplayWidget(NodeWidget):
    """
    Simple Tree widget to show data and their dependencies in the node data.
    """

    def __init__(self, node: Optional[Node] = None):
        self.icon = get_dataColumnsIcon()
        super().__init__(embedWidgetClass=DataSelectionWidget)
        assert self.widget is not None
        self.optSetters = {
            'selectedData': self.setSelected,
        }
        self.optGetters = {
            'selectedData': self.getSelected,
        }

        self.widget.dataSelectionMade.connect(
            lambda x: self.signalOption('selectedData'))

    def setSelected(self, vals: Sequence[str]) -> None:
        assert self.widget is not None
        self.widget.setSelectedData(vals)
        self._updateOptions(vals)

    def getSelected(self) -> List[str]:
        assert self.widget is not None
        return self.widget.getSelectedData()

    def setData(self, structure: DataDictBase,
                shapes: Dict[str, Tuple[int, ...]], _: Any) -> None:
        assert self.widget is not None
        self.widget.setData(structure, shapes)

    def setShape(self, shapes: Dict[str, Tuple[int, ...]]) -> None:
        assert self.widget is not None
        self.widget.setShape(shapes)

    def _updateOptions(self, selected: Sequence[str]) -> None:
        assert self.widget is not None
        ds = self.widget._dataStructure
        for n, w in self.widget.dataItems.items():
            if selected != [] and ds[n]['axes'] != ds[selected[0]]['axes']:
                self.widget.setItemEnabled(n, False)
            else:
                self.widget.setItemEnabled(n, True)


class DataSelector(Node):
    """
    This node allows extracting data from datasets. The fields specified by
    ``selectedData`` and their axes are kept, the rest is discarded.
    All selected data fields must be compatible in the sense that they have the
    same axes (also in the same order).
    The utility of this node is that afterwards data can safely be processed
    together, as the structure of all remaining fields is shared.

    Properties of this node:
    :selectedData: list of strings with compatible dependents.
    """

    # TODO: allow the user to control dtypes.

    nodeName = "DataSelector"
    uiClass = DataDisplayWidget

    force_numerical_data = True

    def __init__(self, name: str):
        super().__init__(name)

        self._dataStructure = None
        self.selectedData = []  # type: ignore[misc]

    # Properties

    @property
    def selectedData(self) -> List[str]:
        return self._selectedData

    @selectedData.setter  # type: ignore[misc]
    @updateOption('selectedData')
    def selectedData(self, val: List[str]) -> None:
        if isinstance(val, str):
            val = [val]
        self._selectedData = val

    # Data processing

    def validateOptions(self, data: DataDictBase) -> bool:
        """
        Validations performed:
        * only compatible data fields can be selected.
        """
        if data is None:
            return True

        for elt in self.selectedData:
            if elt not in data:
                self.logger().warning(
                    f'Did not find selected data {elt} in data. '
                    f'Clearing the selection.'
                )
                self._selectedData = []

        if len(self.selectedData) > 0:
            allowed_axes = data.axes(self.selectedData[0])
            for d in self.selectedData:
                if data.axes(d) != allowed_axes:
                    self.logger().error(
                        f'Datasets {self.selectedData[0]} '
                        f'(with axes {allowed_axes}) '
                        f'and {d}(with axes {data.axes(d)}) are not compatible '
                        f'and cannot be selected simultaneously.'
                        )
                    return False
        return True

    def _reduceData(self, data: Optional[DataDictBase]) -> Optional[DataDictBase]:
        if data is None:
            return None
        if isinstance(self.selectedData, str):
            dnames = [self.selectedData]
        else:
            dnames = self.selectedData
        if len(self.selectedData) == 0:
            return None

        ret = data.extract(dnames)
        if self.force_numerical_data:
            for d, _ in ret.data_items():
                d_data_vals = ret.data_vals(d)
                dt = num.largest_numtype(d_data_vals,
                                         include_integers=False)
                if dt is not None:
                    ret[d]['values'] = ret[d]['values'].astype(dt)
                else:
                    return None

        return ret

    def process(self, dataIn: Optional[DataDictBase] = None) -> Optional[Dict[str, Any]]:
        data = super().process(dataIn=dataIn)
        if data is None:
            return None
        data = data['dataOut']

        # this is the actual operation of the node
        data = self._reduceData(data)
        if data is None:
            return None

        # it is possible at this stage that we have data in DataDictBase format
        # which we cannot process further down the line.
        # But after extraction of compatible date we can now convert.
        if isinstance(data, DataDictBase):
            data = DataDict(**data)
            if not data.validate():
                return None

        return dict(dataOut=data)

    # Methods for GUI interaction

    def setupUi(self) -> None:
        super().setupUi()
        assert self.ui is not None
        self.newDataStructure.connect(self.ui.setData)
        self.dataShapesChanged.connect(self.ui.setShape)




from typing import Sequence, Optional, Dict

import numpy as np

from plottr.node import Node, NodeWidget, updateOption
from plottr.gui.widgets import AxisSelector
from plottr.data.datadict import DataDictBase, MeshgridDataDict


class SubtractAverageWidget(NodeWidget):

    def __init__(self, node: Optional[Node] = None):
        super().__init__(node=node, embedWidgetClass=AxisSelector)
        if self.widget is None:
            raise RuntimeError

        self.optSetters = {
            'averagingAxis': self.setAvgAxis,
        }

        self.optGetters = {
            'averagingAxis': self.getAvgAxis,
        }

        self.combo = self.widget.elements['Axis']
        self.combo.connectNode(self.node)
        self.combo.dimensionSelected.connect(
            lambda x: self.signalOption('averagingAxis')
        )

    def setAvgAxis(self, val: str) -> None:
        self.combo.setCurrentText(val)

    def getAvgAxis(self) -> str:
        return self.combo.currentText()


class SubtractAverage(Node):
    useUi = True
    uiClass = SubtractAverageWidget

    def __init__(self, name: str):
        super().__init__(name)
        self._averagingAxis: Optional[str] = None

    @property
    def averagingAxis(self) -> Optional[str]:
        return self._averagingAxis

    @averagingAxis.setter  # type: ignore[misc]
    @updateOption('averagingAxis')
    def averagingAxis(self, value: str) -> None:
        self._averagingAxis = value

    def process(self, dataIn: Optional[DataDictBase] = None) -> Optional[Dict[str, Optional[DataDictBase]]]:
        if super().process(dataIn=dataIn) is None:
            return None
        assert dataIn is not None
        assert self.dataAxes is not None
        data = dataIn.copy()
        if self._averagingAxis in self.dataAxes and \
                self.dataType == MeshgridDataDict:
            axidx = self.dataAxes.index(self._averagingAxis)
            for dep in dataIn.dependents():
                data_vals = np.asanyarray(data.data_vals(dep))
                avg = data_vals.mean(axis=axidx, keepdims=True)
                data[dep]['values'] -= avg

        return dict(dataOut=data)





import os
from enum import Enum
from typing import List, Any, Optional, Dict, Sequence, Union
from pprint import pprint

from plottr import QtCore, QtGui, QtWidgets, Slot, Signal
from plottr.data.datadict import DataDict


def findFilesByExtension(path: str, extensions: Sequence[str]) -> List[str]:
    ret: List[str] = []
    contents = os.listdir(path)
    for c in sorted(contents):
        abspath = os.path.abspath(os.path.join(path, c))
        if os.path.isdir(abspath):
            ret_ = findFilesByExtension(abspath, extensions)
            if len(ret_) > 0:
                ret += ret_
        else:
            if os.path.splitext(c)[-1] in extensions:
                ret += [abspath]
    return ret


class DataFileContent(QtWidgets.QTreeWidget):

    #: Signal(str) -- Emitted when the user requests a plot for datadict
    #: Arguments:
    #:   - name of the group within the currently selected file
    plotRequested = Signal(str)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        self.data: Dict[str, DataDict] = {}
        self.groupItems: List[QtWidgets.QTreeWidgetItem] = []
        self.selectedGroup = None

        self.dataPopup = QtWidgets.QMenu('Data actions', self)
        self.plotAction = self.dataPopup.addAction("Plot")
        self.plotAction.triggered.connect(self.onPlotActionTriggered)


    @Slot(object)
    def setData(self, data: Dict[str, DataDict]) -> None:
        """Set the data to display."""
        self.clear()
        self.data = {}
        self.groupItems = []

        for grpName, grpData in data.items():
            self.data[grpName] = data[grpName]
            grpItem = QtWidgets.QTreeWidgetItem(self, [grpName])
            self.groupItems.append(grpItem)
            self.addTopLevelItem(grpItem)
            dataParent = QtWidgets.QTreeWidgetItem(grpItem, ['[DATA]'])
            metaParent = QtWidgets.QTreeWidgetItem(grpItem, ['[META]'])

            for dn, dv in grpData.data_items():
                label = grpData.label(dn)
                assert label is not None
                vals = [label, str(grpData.meta_val('shape', dn))]

                if dn in grpData.dependents():
                    vals.append(f'Data (depends on {str(tuple(grpData.axes(dn)))[1:]}')
                else:
                    vals.append('Data (independent)')
                ditem = QtWidgets.QTreeWidgetItem(dataParent, vals)

                for mn, mv in grpData.meta_items(dn):
                    vals = [mn, str(mv)]
                    _ = QtWidgets.QTreeWidgetItem(ditem, vals)

            for mn, mv in grpData.meta_items():
                vals = [mn, str(mv)]
                _ = QtWidgets.QTreeWidgetItem(metaParent, vals)

            grpItem.setExpanded(True)
            dataParent.setExpanded(True)

        for i in range(self.columnCount()-1):
            self.resizeColumnToContents(i)

    @Slot(QtCore.QPoint)
    def onCustomContextMenuRequested(self, pos: QtCore.QPoint) -> None:
        item = self.itemAt(pos)
        if item not in self.groupItems:
            return

        self.selectedGroup = item.text(0)
        self.plotAction.setText(f"Plot '{item.text(0)}'")
        self.dataPopup.exec(self.mapToGlobal(pos))

    @Slot()
    def onPlotActionTriggered(self) -> None:
        self.plotRequested.emit(self.selectedGroup)


class DataFileList(QtWidgets.QTreeWidget):
    """A Tree Widget that displays all data files that are in a certain
    base directory. All subfolders are monitored.
    """

    fileExtensions = ['.ddh5']

    #: Signal(str) -- emitted when a data file is selected
    #: Arguments:
    #:   - the absolute path of the data file
    dataFileSelected = Signal(str)

    #: Signal(list) -- emitted when new files have been found
    newDataFilesFound = Signal(list)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        self.files: List[str] = []
        self.path: Optional[str] = None

    @staticmethod
    def finditem(parent: Union["DataFileList", QtWidgets.QTreeWidgetItem], name: str) -> Optional[QtWidgets.QTreeWidgetItem]:
        if isinstance(parent, DataFileList):
            existingItems = [parent.topLevelItem(i) for i in
                             range(parent.topLevelItemCount())]
        else:
            existingItems = [parent.child(i) for i in
                             range(parent.childCount())]

        item = None
        for item_ in existingItems:
            if item_.text(0) == name:
                item = item_
                break
        return item

    def itemPath(self, item: QtWidgets.QTreeWidgetItem) -> str:
        def buildPath(i: Optional[QtWidgets.QTreeWidgetItem], suffix: str = '') -> str:
            if i is None:
                return suffix
            newSuffix = i.text(0)
            if suffix != '':
                newSuffix += os.path.sep+suffix
            return buildPath(i.parent(), suffix=newSuffix)
        assert self.path is not None
        return os.path.join(self.path, buildPath(item))

    def findItemByPath(self, path: str) -> Optional[QtWidgets.QTreeWidgetItem]:
        assert self.path is not None
        path = path[len(self.path) + len(os.path.sep):]
        pathList = path.split(os.path.sep)
        parent: Union["DataFileList", QtWidgets.QTreeWidgetItem] = self
        for p in pathList:
            new_parent = self.finditem(parent, p)
            if new_parent is None:
                return None
            else:
                parent = new_parent
        assert isinstance(parent, QtWidgets.QTreeWidgetItem)
        return parent

    def addItemByPath(self, path: str) -> None:
        assert self.path is not None
        path = path[len(self.path)+len(os.path.sep):]
        pathList = path.split(os.path.sep)

        def add(parent: Union["DataFileList", QtWidgets.QTreeWidgetItem], name: str) -> Union["DataFileList", QtWidgets.QTreeWidgetItem]:
            item = self.finditem(parent, name)
            if item is None:
                item = QtWidgets.QTreeWidgetItem(parent, [name])
                if os.path.splitext(name)[-1] in self.fileExtensions:
                    fnt = QtGui.QFont()
                    item.setFont(0, fnt)
                else:
                    pass
                if isinstance(parent, DataFileList):
                    parent.addTopLevelItem(item)
                else:
                    parent.addChild(item)
            return item

        parent: Union["DataFileList", QtWidgets.QTreeWidgetItem] = self
        for p in pathList:
            parent = add(parent, p)

    def removeItemByPath(self, path: str) -> None:

        def remove(i: QtWidgets.QTreeWidgetItem) -> None:
            parent = i.parent()
            if isinstance(parent, DataFileList):
                idx = parent.indexOfTopLevelItem(i)
                parent.takeTopLevelItem(idx)
            elif isinstance(parent, QtWidgets.QTreeWidgetItem):
                parent.removeChild(i)
                if parent.childCount() == 0:
                    remove(parent)

        item = self.findItemByPath(path)
        if item is None:
            return
        remove(item)

    def loadFromPath(self, path: str, emitNew: bool = False) -> None:
        self.path = path
        files = findFilesByExtension(path, self.fileExtensions)
        newFiles = [f for f in files if f not in self.files]
        removedFiles = [f for f in self.files if f not in files]

        for f in newFiles:
            self.addItemByPath(f)

        for f in removedFiles:
            self.removeItemByPath(f)

        self.files = files
        if len(newFiles) > 0 and emitNew:
            self.newDataFilesFound.emit(newFiles)

    @Slot()
    def processSelection(self) -> None:
        selected = self.selectedItems()
        if len(selected) == 0:
            return
        nameAndExt = os.path.splitext(selected[0].text(0))
        if nameAndExt[-1] in self.fileExtensions:
            path = self.itemPath(selected[0])
            self.dataFileSelected.emit(path)


class MonitorToolBar(QtWidgets.QToolBar):
    pass



"""Tools for automatic plotting with ``pyqtgraph``.

Contains a ``pyqtgraph``-specific implementation of :class:`.plot.base.FigureMaker`
(:class:`.FigureMaker`) and :class:`.plot.base.PlotWidget` (:class:`.AutoPlot`).
``FigureMaker`` can be used to quickly create a figure in a mostly automatic fashion
from known data.
For embedding in GUIs (such as plottr applications), ``AutoPlot`` is the main
object for plotting data automatically using ``pyqtgraph``.
"""

import logging
from dataclasses import dataclass
from typing import List, Optional, Any

import numpy as np
from pyqtgraph import mkPen

from plottr import QtWidgets, QtCore, Signal, Slot, \
    config_entry as getcfg
from plottr.data.datadict import DataDictBase
from .plots import Plot, PlotWithColorbar, PlotBase
from ..base import AutoFigureMaker as BaseFM, PlotDataType, \
    PlotItem, ComplexRepresentation, determinePlotDataType, \
    PlotWidgetContainer, PlotWidget

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class FigureWidget(QtWidgets.QWidget):
    """Widget that contains all plots generated by :class:`.FigureMaker`.

    Widget has a vertical layout, and plots can be added in a single column.
    """

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        """Constructor for :class:`.FigureMakerWidget`.

        :param parent: parent widget.
        """
        super().__init__(parent=parent)

        self.subPlots: List[PlotBase] = []

        self.title = QtWidgets.QLabel(parent=self)
        self.title.setAlignment(QtCore.Qt.AlignHCenter)

        self.split = QtWidgets.QSplitter(parent=self)
        self.split.setOrientation(QtCore.Qt.Vertical)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        layout.addWidget(self.title)
        layout.addWidget(self.split)
        self.setLayout(layout)

        self.setTitle('')

    def addPlot(self, plot: PlotBase) -> None:
        """Add a :class:`.PlotBase` widget.

        :param plot: plot widget
        :param title: title of the plot
        """
        self.split.addWidget(plot)
        self.subPlots.append(plot)

    def clearAllPlots(self) -> None:
        """Clear all plot contents."""
        for p in self.subPlots:
            p.clearPlot()

    def deleteAllPlots(self) -> None:
        """Delete all subplot widgets."""
        for p in self.subPlots:
            p.deleteLater()
        self.subPlots = []

    def setTitle(self, title: str = '') -> None:
        self.title.setText(title)
        if len(title.strip()) == 0:
            self.title.setVisible(False)
        else:
            self.title.setVisible(True)


class FigureMaker(BaseFM):
    """``pyqtgraph`` implementation for :class:`.AutoFigureMaker`.

    Please see the base class for more general information.
    """

    # TODO: make scrollable when many figures (set min size)?
    # TODO: check for valid plot data

    def __init__(self, widget: Optional[FigureWidget] = None,
                 clearWidget: bool = True,
                 parentWidget: Optional[QtWidgets.QWidget] = None):
        """Constructor for :class:`.FigureMaker`.

        :param widget: a widget produced by an earlier use of this class
            may be supplied. This widget will then be used for generating the
            plot and avoids deleting and re-creating the widget if no structural
            changes to the plots is required (important case is live-plotting of
            data that keeps updating).
        :param clearWidget:
            if ``True`` force re-creation of all plot elements.
        :param parentWidget:
            parent of the main widget of FigureMaker.
        """
        super().__init__()

        self.widget: FigureWidget
        self.clearWidget = clearWidget

        if widget is None:
            self.widget = FigureWidget(parent=parentWidget)
        else:
            self.widget = widget

    # re-implementing to get correct type annotation.
    def __enter__(self) -> "FigureMaker":
        return self

    def subPlotFromId(self, subPlotId: int) -> PlotBase:
        """Get SubPlot from ID."""
        subPlots = self.subPlots[subPlotId].axes
        assert isinstance(subPlots, list) and len(subPlots) > 0 and \
               isinstance(subPlots[0], PlotBase)
        return subPlots[0]

    def makeSubPlots(self, nSubPlots: int) -> List[PlotBase]:
        """Create empty subplots in the widgets.

        If ``clearWidget`` was not set to ``True`` in the constructor,
        existing sub plot widgets are cleared, but not deleted and re-created.
        """
        plot: PlotBase
        if self.clearWidget:
            self.widget.deleteAllPlots()
            for i in range(nSubPlots):
                if max(self.dataDimensionsInSubPlot(i).values()) == 1:
                    plot = Plot(self.widget)
                    self.widget.addPlot(plot)
                elif max(self.dataDimensionsInSubPlot(i).values()) == 2:
                    plot = PlotWithColorbar(self.widget)
                    self.widget.addPlot(plot)
        else:
            self.widget.clearAllPlots()

        return self.widget.subPlots

    def formatSubPlot(self, subPlotId: int) -> None:
        """Set labels/legends for the given subplot."""
        if len(self.plotIdsInSubPlot(subPlotId)) == 0:
            return

        labels = self.subPlotLabels(subPlotId)
        subPlot = self.subPlotFromId(subPlotId)

        # label the x axis if there's only one x label
        if isinstance(subPlot, Plot):
            if len(set(labels[0])) == 1:
                subPlot.plot.setLabel("bottom", labels[0][0])

        if isinstance(subPlot, PlotWithColorbar):
            if len(set(labels[0])) == 1:
                subPlot.plot.setLabel("bottom", labels[0][0])

            if len(set(labels[1])) == 1:
                subPlot.plot.setLabel('left', labels[1][0])

            if len(set(labels[2])) == 1:
                subPlot.colorbar.setLabel('left', labels[2][0])

    def plot(self, plotItem: PlotItem) -> None:
        """Plot the given item."""
        if plotItem.plotDataType is PlotDataType.unknown:
            if len(plotItem.data) == 2:
                plotItem.plotDataType = PlotDataType.scatter1d
            elif len(plotItem.data) == 3:
                plotItem.plotDataType = PlotDataType.scatter2d

        if plotItem.plotDataType in [PlotDataType.scatter1d, PlotDataType.line1d]:
            self._1dPlot(plotItem)
        elif plotItem.plotDataType == PlotDataType.grid2d:
            self._colorPlot(plotItem)
        elif plotItem.plotDataType == PlotDataType.scatter2d:
            self._scatterPlot2d(plotItem)
        else:
            raise NotImplementedError('Cannot plot this data.')

    def _1dPlot(self, plotItem: PlotItem) -> None:
        colors = getcfg('main', 'pyqtgraph', 'line_colors', default=['r', 'b', 'g'])
        symbols = getcfg('main', 'pyqtgraph', 'line_symbols', default=['o'])
        symbolSize = getcfg('main', 'pyqtgraph', 'line_symbol_size', default=5)

        subPlot = self.subPlotFromId(plotItem.subPlot)

        assert len(plotItem.data) == 2
        x, y = plotItem.data

        color = colors[self.findPlotIndexInSubPlot(plotItem.id) % len(colors)]
        symbol = symbols[self.findPlotIndexInSubPlot(plotItem.id) % len(symbols)]

        if plotItem.plotDataType == PlotDataType.line1d:
            name = plotItem.labels[-1] if isinstance(plotItem.labels, list) else ''
            return subPlot.plot.plot(x.flatten(), y.flatten(), name=name,
                                     pen=mkPen(color, width=1), symbol=symbol, symbolBrush=color,
                                     symbolPen=None, symbolSize=symbolSize)
        else:
            name = plotItem.labels[-1] if isinstance(plotItem.labels, list) else ''
            return subPlot.plot.plot(x.flatten(), y.flatten(), name=name,
                                     pen=None, symbol=symbol, symbolBrush=color,
                                     symbolPen=None, symbolSize=symbolSize)

    def _colorPlot(self, plotItem: PlotItem) -> None:
        subPlot = self.subPlotFromId(plotItem.subPlot)
        assert isinstance(subPlot, PlotWithColorbar) and len(plotItem.data) == 3
        subPlot.setImage(*plotItem.data)

    def _scatterPlot2d(self, plotItem: PlotItem) -> None:
        subPlot = self.subPlotFromId(plotItem.subPlot)
        assert isinstance(subPlot, PlotWithColorbar) and len(plotItem.data) == 3
        subPlot.setScatter2d(*plotItem.data)


class AutoPlot(PlotWidget):
    """Widget for automatic plotting with pyqtgraph.

    Uses :class:`.FigureMaker` to produce subplots.
    """

    def __init__(self, parent: Optional[PlotWidgetContainer]) -> None:
        """Constructor for the pyqtgraph auto plot widget.

        :param parent: plot widget container
        """
        super().__init__(parent=parent)

        self.fmWidget: Optional[FigureWidget] = None
        self.figConfig: Optional[FigureConfigToolBar] = None
        self.figOptions: FigureOptions = FigureOptions()

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.setLayout(layout)
        self.setMinimumSize(*getcfg('main', 'pyqtgraph', 'minimum_plot_size',
                                    default=(400, 400)))

    def setData(self, data: Optional[DataDictBase]) -> None:
        """Uses :class:`.FigureMaker` to populate the plot(s).

        This method aims to re-use existing plotwidgets if possible -- i.e.,
        when the type of data (generally speaking) has not changed.
        If changes to the data structure are detected, then all subplot
        widgets are re-created from scratch, however.

        :param data: input data
        :return: ``None``
        """
        super().setData(data)
        if self.data is None:
            return

        fmKwargs = {}  # {'widget': self.fmWidget}
        dc = self.dataChanges
        if not dc['dataTypeChanged'] and not dc['dataStructureChanged']:
            fmKwargs['clearWidget'] = False
        else:
            fmKwargs['clearWidget'] = True
        self._plotData(**fmKwargs)

    def _plotData(self, **kwargs: Any) -> None:
        if self.data is None:
            return

        with FigureMaker(parentWidget=self, widget=self.fmWidget,
                         **kwargs) as fm:

            fm.complexRepresentation = self.figOptions.complexRepresentation
            fm.combineTraces = self.figOptions.combineLinePlots

            for dep in self.data.dependents():
                inds = self.data.axes(dep)
                dvals = self.data.data_vals(dep)
                pdt = determinePlotDataType(self.data.extract([dep]))
                plotId = fm.addData(
                    *[np.asanyarray(self.data.data_vals(n)) for n in inds] + [dvals],
                    labels=[str(self.data.label(n)) for n in inds] + [str(self.data.label(dep))],
                    plotDataType=pdt,
                )

        if self.fmWidget is None:
            self.fmWidget = fm.widget
            self.layout().addWidget(self.fmWidget)
            self.figConfig = FigureConfigToolBar(self.figOptions,
                                                 parent=self)
            self.layout().addWidget(self.figConfig)
            self.figConfig.optionsChanged.connect(self._refreshPlot)

        if self.data.has_meta('title'):
            self.fmWidget.setTitle(self.data.meta_val('title'))

    @Slot()
    def _refreshPlot(self) -> None:
        self._plotData()


@dataclass
class FigureOptions:
    """Dataclass that describes the configuration options for the figure."""

    #: whether to plot all 1D traces into a single panel
    combineLinePlots: bool = False

    #: how to represent complex data
    complexRepresentation: ComplexRepresentation = ComplexRepresentation.realAndImag


class FigureConfigToolBar(QtWidgets.QToolBar):
    """Simple toolbar to configure the figure."""

    # TODO: find better config system that generates GUI automatically and
    #   links updates easier.

    #: Signal() -- emitted when options have been changed in the GUI.
    optionsChanged = Signal()

    def __init__(self, options: FigureOptions,
                 parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Constructor.

        :param options: options object. GUI interaction will make changes
            in-place to this object.
        :param parent: parent Widget
        """
        super().__init__(parent)

        self.options = options

        combineLinePlots = self.addAction("Combine 1D")
        combineLinePlots.setCheckable(True)
        combineLinePlots.setChecked(self.options.combineLinePlots)
        combineLinePlots.triggered.connect(
            lambda: self._setOption('combineLinePlots',
                                    combineLinePlots.isChecked())
        )

        complexOptions = QtWidgets.QMenu(parent=self)
        complexGroup = QtWidgets.QActionGroup(complexOptions)
        complexGroup.setExclusive(True)
        for k in ComplexRepresentation:
            a = QtWidgets.QAction(k.label, complexOptions)
            a.setCheckable(True)
            complexGroup.addAction(a)
            complexOptions.addAction(a)
            a.setChecked(k == self.options.complexRepresentation)
        complexGroup.triggered.connect(
            lambda _a: self._setOption('complexRepresentation',
                                       ComplexRepresentation.fromLabel(_a.text()))
        )
        complexButton = QtWidgets.QToolButton()
        complexButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        complexButton.setText('Complex')
        complexButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        complexButton.setMenu(complexOptions)
        self.addWidget(complexButton)

    def _setOption(self, option: str, value: Any) -> None:
        setattr(self.options, option, value)
        self.optionsChanged.emit()



"""Convenience tools for generating ``pyqtgraph`` plots that can
be used in plottr's automatic plotting framework."""

from typing import Optional, Tuple, NoReturn

import numpy as np
import pyqtgraph as pg

from plottr import QtCore, QtWidgets

__all__ = ['PlotBase', 'Plot']


class PlotBase(QtWidgets.QWidget):
    """A simple convenience widget class as container for ``pyqtgraph`` plots.

    The widget contains a layout that contains a ``GraphicsLayoutWidget``.
    This is handy because a plot may contain multiple elements (like an image
    and a colorbar).

    This base class should be inherited to use.
    """

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)

        #: central layout of the widget. only contains a graphics layout.
        layout = QtWidgets.QHBoxLayout(self)
        #: ``pyqtgraph`` graphics layout
        self.graphicsLayout = pg.GraphicsLayoutWidget(self)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)
        layout.addWidget(self.graphicsLayout)

        #: ``pyqtgraph`` plot item
        self.plot: pg.PlotItem = self.graphicsLayout.addPlot()

    def clearPlot(self) -> None:
        """Clear all plot contents (but do not delete plot elements, like axis
        spines, insets, etc).

        To be implemented by inheriting classes."""
        raise NotImplementedError


class Plot(PlotBase):
    """A simple plot with a single ``PlotItem``."""

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        legend = self.plot.addLegend(offset=(5, 5), pen='#999',
                                     brush=(255, 255, 255, 150))
        legend.layout.setContentsMargins(0, 0, 0, 0)
        self.plot.showGrid(True, True)

    def clearPlot(self) -> None:
        """Clear the plot item."""
        self.plot.clear()


class PlotWithColorbar(PlotBase):
    """Plot containing a plot item and a colorbar item.

    Plot is suited for either an image plot (:meth:`.setImage`) or a color
    scatter plot (:meth:`.setScatter2D`).
    The color scale is displayed in an interactive colorbar.
    """
    #: colorbar
    colorbar: pg.ColorBarItem

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)

        cmap = pg.colormap.get('viridis', source='matplotlib')
        self.colorbar: pg.ColorBarItem = pg.ColorBarItem(interactive=True, values=(0, 1),
                                                         cmap=cmap, width=15)
        self.graphicsLayout.addItem(self.colorbar)

        self.img: Optional[pg.ImageItem] = None
        self.scatter: Optional[pg.ScatterPlotItem] = None
        self.scatterZVals: Optional[np.ndarray] = None

    def clearPlot(self) -> None:
        """Clear the content of the plot."""
        self.img = None
        self.scatter = None
        self.scatterZVals = None
        self.plot.clear()
        try:
            self.colorbar.sigLevelsChanged.disconnect(self._colorScatterPoints)
        except TypeError:
            pass

    def setImage(self, x: np.ndarray, y: np.ndarray, z: np.ndarray) -> None:
        """Set data to be plotted as image.

        Clears the plot before creating a new image item that gets places in the
        plot and linked to the colorscale.

        :param x: x coordinates (as 2D meshgrid)
        :param y: y coordinates (as 2D meshgrid)
        :param z: data values (as 2D meshgrid)
        :return: None
        """
        self.clearPlot()

        self.img = pg.ImageItem()
        self.plot.addItem(self.img)
        self.img.setImage(z)
        self.img.setRect(QtCore.QRectF(x.min(), y.min(), x.max() - x.min(), y.max() - y.min()))

        self.colorbar.setImageItem(self.img)
        self.colorbar.rounding = (z.max() - z.min()) * 1e-2
        self.colorbar.setLevels((z.min(), z.max()))

    def setScatter2d(self, x: np.ndarray, y: np.ndarray, z: np.ndarray) -> None:
        """Set data to be plotted as image.

        Clears the plot before creating a new scatter item (based on flattened
        input data) that gets placed in the plot and linked to the colorscale.

        :param x: x coordinates
        :param y: y coordinates
        :param z: data values
        :return: None
        """
        self.clearPlot()

        self.scatter = pg.ScatterPlotItem()
        self.scatter.setData(x=x.flatten(), y=y.flatten(), symbol='o', size=8)
        self.plot.addItem(self.scatter)
        self.scatterZVals = z.flatten()

        self.colorbar.setLevels((z.min(), z.max()))
        self.colorbar.rounding = (z.max() - z.min()) * 1e-2
        self._colorScatterPoints(self.colorbar)

        self.colorbar.sigLevelsChanged.connect(self._colorScatterPoints)

    # TODO: this seems crazy slow.
    def _colorScatterPoints(self, cbar: pg.ColorBarItem) -> None:
        if self.scatter is not None and self.scatterZVals is not None:
            z_norm = self._normalizeColors(self.scatterZVals, cbar.levels())
            colors = self.colorbar.cmap.mapToQColor(z_norm)
            self.scatter.setBrush(colors)

    def _normalizeColors(self, z: np.ndarray, levels: Tuple[float, float]) -> np.ndarray:
        scale = levels[1] - levels[0]
        if scale > 0:
            return (z - levels[0]) / scale
        else:
            return np.ones(z.size) * 0.5



"""
``plottr.plot.mpl.plotting`` -- Plotting tools (mostly used in Autoplot)
"""

from enum import Enum, auto, unique
from typing import Any, Optional, Tuple, Union

import numpy as np
from matplotlib import colors, rcParams
from matplotlib.axes import Axes
from matplotlib.image import AxesImage

from plottr.utils import num
from plottr.utils.num import centers2edges_2d, interp_meshgrid_2d

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'


@unique
class PlotType(Enum):
    """Plot types currently supported in Autoplot."""

    #: no plot defined
    empty = auto()

    #: a single 1D line/scatter plot per panel
    singletraces = auto()

    #: multiple 1D lines/scatter plots per panel
    multitraces = auto()

    #: image plot of 2D data
    image = auto()

    #: colormesh plot of 2D data
    colormesh = auto()

    #: 2D scatter plot
    scatter2d = auto()


class SymmetricNorm(colors.Normalize):
    """Color norm that's symmetric and linear around a center value."""

    def __init__(self, vmin: Optional[float] = None,
                 vmax: Optional[float] = None,
                 vcenter: float = 0,
                 clip: bool = False):
        super().__init__(vmin, vmax, clip)
        self.vcenter = vcenter

    def __call__(self, value: float, clip: Optional[bool] = None) -> np.ma.core.MaskedArray:
        vlim = max(abs(self.vmin - self.vcenter), abs(self.vmax - self.vcenter))
        self.vmax: float = vlim + self.vcenter
        self.vmin: float = -vlim + self.vcenter
        return super().__call__(value, clip)


# 2D plots
def colorplot2d(ax: Axes,
                x: Union[np.ndarray, np.ma.MaskedArray],
                y: Union[np.ndarray, np.ma.MaskedArray],
                z: Union[np.ndarray, np.ma.MaskedArray],
                plotType: PlotType = PlotType.image,
                axLabels: Tuple[Optional[str], Optional[str], Optional[str]] = ('', '', ''),
                **kw: Any) -> Optional[AxesImage]:
    """make a 2d colorplot. what plot is made, depends on `plotType`.
    Any of the 2d plot types in :class:`PlotType` works.

    :param ax: matplotlib subPlots to plot in
    :param x: x coordinates (meshgrid)
    :param y: y coordinates (meshgrid)
    :param z: z data
    :param plotType: the plot type
    :param axLabels: labels for the x, y subPlots, and the colorbar.

    all keywords are passed to the actual plotting functions, depending on the ``plotType``:

    - :attr:`PlotType.image` --
        :func:`plotImage`
    - :attr:`PlotType.colormesh` --
        :func:`ppcolormesh_from_meshgrid`
    - :attr:`PlotType.scatter2d` --
        matplotlib's `scatter`
    """
    cmap = kw.pop('cmap', rcParams['image.cmap'])

    # first we need to check if our grid can be plotted nicely.
    if plotType in [PlotType.image, PlotType.colormesh]:
        x = x.astype(float)
        y = y.astype(float)
        z = z.astype(float)

        # first check if we need to fill some masked values in
        if isinstance(x, np.ma.MaskedArray) and np.ma.is_masked(x):
            x = x.filled(np.nan)
        if isinstance(y, np.ma.MaskedArray) and np.ma.is_masked(y):
            y = y.filled(np.nan)
        if isinstance(z, np.ma.MaskedArray) and np.ma.is_masked(z):
            z = z.filled(np.nan)

        # next: try some surgery, if possible
        if np.all(num.is_invalid(x)) or np.all(num.is_invalid(y)):
            return None
        if np.any(np.isnan(x)) or np.any(np.isnan(y)):
            x, y = interp_meshgrid_2d(x, y)
        if np.any(num.is_invalid(x)) or np.any(num.is_invalid(y)):
            x, y, z = num.crop2d(x, y, z)

        # next, check if the resulting grids are even still plottable
        for g in x, y, z:
            if g.size == 0:
                return None
            elif len(g.shape) < 2:
                return None

            # special case: if we have a single line, a pcolor-type plot won't work.
            elif min(g.shape) < 2:
                plotType = PlotType.scatter2d

    if plotType is PlotType.image:
        im = plotImage(ax, x, y, z, cmap=cmap, **kw)
    elif plotType is PlotType.colormesh:
        im = ppcolormesh_from_meshgrid(ax, x, y, z, cmap=cmap, **kw)
    elif plotType is PlotType.scatter2d:
        im = ax.scatter(x.ravel(), y.ravel(), c=z.ravel(), cmap=cmap, **kw)
    else:
        im = None

    if im is None:
        return None

    ax.set_xlabel(axLabels[0])
    ax.set_ylabel(axLabels[1])
    return im


def ppcolormesh_from_meshgrid(ax: Axes, x: np.ndarray, y: np.ndarray,
                              z: np.ndarray, **kw: Any) -> Union[AxesImage, None]:
    r"""Plot a pcolormesh with some reasonable defaults.
    Input are the corresponding arrays from a 2D ``MeshgridDataDict``.

    Will attempt to fix missing points in the coordinates.

    :param ax: subPlots to plot the colormesh into.
    :param x: x component of the meshgrid coordinates
    :param y: y component of the meshgrid coordinates
    :param z: data values
    :returns: the image returned by `pcolormesh`.

    Keywords are passed on to `pcolormesh`.
    """
    # the meshgrid we have describes coordinates, but for plotting
    # with pcolormesh we need vertices.
    try:
        x = centers2edges_2d(x)
        y = centers2edges_2d(y)
    except:
        return None

    im = ax.pcolormesh(x, y, z, **kw)
    ax.set_xlim(x.min(), x.max())
    ax.set_ylim(y.min(), y.max())
    return im


def plotImage(ax: Axes, x: np.ndarray, y: np.ndarray,
              z: np.ndarray, **kw: Any) -> AxesImage:
    """Plot 2d meshgrid data as image.

    :param ax: matplotlib subPlots to plot the image in.
    :param x: x coordinates (as meshgrid)
    :param y: y coordinates
    :param z: z values
    :returns: the image object returned by `imshow`

    All keywords are passed to `imshow`.
    """
    ax.grid(False)
    x0, x1 = x.min(), x.max()
    y0, y1 = y.min(), y.max()

    extentx = [x0, x1]
    if x0 > x1:
        extentx = extentx[::-1]
    if x0 == x1:
        extentx = [x0, x0 + 1]
    extenty = [y0, y1]
    if y0 > y1:
        extenty = extenty[::-1]
    if y0 == y1:
        extenty = [y0, y0 + 1]
    extent = tuple(extentx + extenty)

    if x.shape[0] > 1:
        # in image mode we have to be a little careful:
        # if the x/y subPlots are specified with decreasing values we need to
        # flip the image. otherwise we'll end up with an axis that has the
        # opposite ordering from the data.
        z = z if x[0, 0] < x[1, 0] else z[::-1, :]

    if y.shape[1] > 1:
        z = z if y[0, 0] < y[0, 1] else z[:, ::-1]

    im = ax.imshow(z.T, aspect='auto', origin='lower',
                   extent=extent, **kw)
    return im



"""
node.py

Contains the base class for Nodes.
"""
import traceback
from logging import Logger

from functools import wraps
from typing import Any, Union, Tuple, Dict, Optional, Type, List, Callable, TypeVar

from .. import NodeBase
from .. import QtGui, QtCore, Signal, Slot, QtWidgets
from ..data.datadict import DataDictBase, MeshgridDataDict
from .. import log

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'


R = TypeVar('R', bound="Node")
S = TypeVar('S')
T = TypeVar('T')


def updateOption(optName: Optional[str] = None) -> Callable[[Callable[[R, S], T]], Callable[[R, S], T]]:
    """Decorator for property setters that are handy for user options.

    Property setters in nodes that are decorated with this will do two things:
    * call ``Node.update``, in order to update the flowchart.
    * if there is a UI, we call the matching ``optSetter`` function.

    :param optName: name of the property.
    """

    def decorator(func: Callable[[R, S], T]) -> Callable[[R, S], T]:
        @wraps(func)
        def wrap(self: R, val: S) -> T:
            ret = func(self, val)
            if optName is not None and self.ui is not None and \
                    optName in self.ui.optSetters:
                self.ui.optSetters[optName](val)
            self.update(self.signalUpdate)
            return ret

        return wrap

    return decorator


U = TypeVar('U', bound="NodeWidget")
V = TypeVar('V',)


def updateGuiFromNode(func: Callable[..., V]) -> Callable[..., V]:
    """
    Decorator for the UI to set an internal flag to during execution of
    the wrapped function. Prevents recursive updating (i.e., if
    the node sends a new option value to the UI for updating, the UI
    will then `not` notify the node back after making the update).
    """

    @wraps(func)
    def wrap(self: U, *arg: Any, **kw: Any) -> V:
        self._emitGuiChange = False
        ret = func(self, *arg, **kw)
        self._emitGuiChange = True
        return ret

    return wrap


updateGuiQuietly = updateGuiFromNode

W = TypeVar('W')


def emitGuiUpdate(signalName: str) -> Callable[[Callable[..., Any]], Callable[..., None]]:
    """
    Decorator for UI functions to emit the signal ``signalName``
    (given as argument the decorator), with the return of the wrapped function.

    Signal is only emitted if the flag controlled by ``updateGuiFromNode``
    is not ``True``, i.e., if the option change was `not` caused by a
    function decorated with ``updateGuiFromNode``.

    :param signalName: name of the signal to emit.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., None]:
        @wraps(func)
        def wrap(self: W, *arg: Any, **kw: Any) -> None:
            ret = func(self, *arg, **kw)
            emit = getattr(self, '_emitGuiChange', True)
            if emit:
                sig = getattr(self, signalName)
                sig.emit(ret)

        return wrap

    return decorator


# TODO: should we add a list of options to the class?
#   that would allow programmatic syncing from node to widget, for instance.
class Node(NodeBase):
    """Base class of the Node we use for plotter.

    This class inherits from ``pyqtgraph``'s Node, and adds a few additional
    tools, and some defaults.
    """

    #: Name of the node. used in the flowchart node library.
    nodeName = 'Node'

    #: Default terminals: one input and one output.
    terminals = {
        'dataIn': {'io': 'in'},
        'dataOut': {'io': 'out'},
    }

    #: UI node widget class. If not None, and ``useUi`` is ``True``, an
    #: instance of the widget is created, and signal/slots are connected.
    uiClass: Optional[Type["NodeWidget"]] = None

    #: Whether or not to automatically set up a UI widget.
    useUi = True

    #: Whether the ui should be visible by default
    uiVisibleByDefault = False

    #: A signal to notify the UI of option changes
    #: arguments is a dictionary of options and new values.
    optionChangeNotification = Signal(dict)

    #: signal emitted when available data axes change
    #: emits a the list of names of new axes
    dataAxesChanged = Signal(list)

    #: signal emitted when available dependents change
    #: emits a the list of names of new dependents
    dataDependentsChanged = Signal(list)

    #: signal emitted when any available data fields change (dep. and indep.)
    #: emits a the list of names of new axes
    dataFieldsChanged = Signal(list)

    #: signal emitted when data type changes
    dataTypeChanged = Signal(object)

    #: signal emitted when data structure changes (fields, or dtype)
    dataStructureChanged = Signal(object)

    #: signal emitted when data shapes change
    dataShapesChanged = Signal(dict)

    #: when data structure changes, emits (structure, shapes, type)
    newDataStructure = Signal(object, object, object)

    #: developer flag for whether we actually want to raise of use the logging
    #: system
    _raiseExceptions = False

    def __init__(self, name: str):
        """Create a new instance of the Node.

        :param name: name of the instance.
        """
        super().__init__(name, terminals=self.__class__.terminals)

        self.signalUpdate = True
        self.dataAxes: Optional[List[str]] = None
        self.dataDependents: Optional[List[str]] = None
        self.dataType: Optional[Type[DataDictBase]] = None
        self.dataShapes: Optional[Dict[str, Tuple[int, ...]]] = None
        self.dataStructure: Optional[DataDictBase] = None

        if self.useUi and self.__class__.uiClass is not None:
            self.ui: Optional["NodeWidget"] = self.__class__.uiClass(node=self)
            self.setupUi()
        else:
            self.ui = None

    def setupUi(self) -> None:
        """ setting up the UI widget.

        Gets called automatically in the node initialization.
        Automatically connect the UIs methods to signal option values.

        Inheriting classes can use this method to do additional setup of the
        UI widget (like connecting additional signals/slots between node and
        node widget).
        """
        assert self.ui is not None
        self.ui.optionToNode.connect(self.setOption)
        self.ui.allOptionsToNode.connect(self.setOptions)
        self.optionChangeNotification.connect(self.ui.setOptionsFromNode)

    def ctrlWidget(self) -> Union[QtWidgets.QWidget, None]:
        """Returns the node widget, if it exists.
        """
        return self.ui

    def setOption(self, nameAndVal: Tuple[str, Any]) -> None:
        """Set an option.

        name is the name of the property, not the string used for referencing
        (which could in principle be different).

        :param nameAndVal: tuple of option name and new value
        """
        name, val = nameAndVal
        setattr(self, name, val)

    def setOptions(self, opts: Dict[str, Any]) -> None:
        """Set multiple options.

        :param opts: a dictionary of property name : value pairs.
        """
        for opt, val in opts.items():
            setattr(self, opt, val)

    def update(self, signal: bool = True) -> None:
        super().update(signal=signal)
        if Node._raiseExceptions and self.exception is not None:
            raise self.exception[1]
        elif self.exception is not None:
            e = self.exception
            err = f'EXCEPTION RAISED: {e[0]}: {e[1]}\n'
            for t in traceback.format_tb(e[2]):
                err += f' -> {t}\n'
            self.logger().error(err)

    def logger(self) -> Logger:
        """Get a logger for this node

        :return: logger with a name that can be traced back easily to this node.
        """
        name = f"{self.__module__}.{self.__class__.__name__}.{self.name()}"
        logger = log.getLogger(name)
        logger.setLevel(log.LEVEL)
        return logger

    def validateOptions(self, data: DataDictBase) -> bool:
        """Validate the user options

        Does nothing in this base implementation. Can be reimplemented by any
        inheriting class.

        :param data: the data to verify the options against.
        """
        return True

    # TODO: should think about nodes with multiple inputs -- how would this look then?
    # FIXME: return should only be Optional[Dict[str, DataDictBase]]
    def process(self, dataIn: Optional[DataDictBase]=None) -> Optional[Dict[str, Optional[DataDictBase]]]:
        if dataIn is None:
            return None

        if not isinstance(dataIn, DataDictBase):
            raise ValueError('Unsupported data format provided.')

        _axesChanged = False
        _fieldsChanged = False
        _typeChanged = False
        _structChanged = False
        _shapesChanged = False
        _depsChanged = False

        dtype = type(dataIn)
        daxes = dataIn.axes()
        ddeps = dataIn.dependents()
        dshapes = dataIn.shapes()
        dstruct = dataIn.structure(add_shape=False)

        if None in [self.dataAxes, self.dataDependents, self.dataType, self.dataShapes]:
            _axesChanged = True
            _fieldsChanged = True
            _typeChanged = True
            _structChanged = True
            _shapesChanged = True
            _depsChanged = True

        else:
            if daxes != self.dataAxes:
                _fieldsChanged = True
                _structChanged = True
                _axesChanged = True

            if ddeps != self.dataDependents:
                _fieldsChanged = True
                _structChanged = True
                _depsChanged = True

            if dtype != self.dataType:
                _typeChanged = True
                _structChanged = True

            if dshapes != self.dataShapes:
                _shapesChanged = True

        self.dataAxes = daxes
        self.dataDependents = ddeps
        self.dataType = dtype
        self.dataShapes = dshapes
        self.dataStructure = dstruct

        if _axesChanged:
            self.dataAxesChanged.emit(daxes)

        if _depsChanged:
            self.dataDependentsChanged.emit(ddeps)

        if _fieldsChanged:
            self.dataFieldsChanged.emit(daxes + ddeps)

        if _typeChanged:
            self.dataTypeChanged.emit(dtype)

        if _structChanged:
            self.dataStructureChanged.emit(self.dataStructure)
            self.newDataStructure.emit(
                self.dataStructure, self.dataShapes, self.dataType)

        if _shapesChanged and not _structChanged:
            self.dataShapesChanged.emit(dshapes)

        if not self.validateOptions(dataIn):
            self.logger().debug("Option validation not passed")
            return None

        return dict(dataOut=dataIn)


class NodeWidget(QtWidgets.QWidget):
    """
    Base class for Node control widgets.

    For the widget class to set up communication with the Node automatically,
    make sure to set :attr:`plottr.node.node.NodeWidget.optGetters` and
    :attr:`plottr.node.node.NodeWidget.optSetters` for a widget class.
    """

    #: icon for this node
    icon: Optional[QtGui.QIcon] = None

    #: preferred location of the widget when used as dock widget
    preferredDockWidgetArea = QtCore.Qt.LeftDockWidgetArea

    #: signal (args: object)) to emit to notify the node of a (changed)
    #: user option.
    optionToNode = Signal(object)

    #: signal (args: (object)) all options to the node.
    allOptionsToNode = Signal(object)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None,
                 embedWidgetClass: Optional[Type[QtWidgets.QWidget]] = None,
                 node: Optional[Node] = None):
        super().__init__(parent)

        self.optGetters: Dict[str, Callable[[], Any]] = {}
        self.optSetters: Dict[str, Callable[[Any], None]] = {}
        self.node = node

        self._emitGuiChange = True

        self.widget: Optional[QtWidgets.QWidget] = None

        if embedWidgetClass is not None:
            layout = QtWidgets.QVBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            self.widget = embedWidgetClass()
            layout.addWidget(self.widget)
            self.setLayout(layout)

    def getAllOptions(self) -> Dict[str, Any]:
        """Return all options as a dictionary"""
        ret = {}
        for n, f in self.optGetters.items():
            ret[n] = f()

        return ret

    @updateGuiFromNode
    def setOptionFromNode(self, opt: str, value: Any) -> None:
        """Set an option from the node

        Calls the set function specified in the class' ``optSetters``.
        Decorated with ``@updateGuiFromNode``.

        :param opt: name of the option
        :param value: value to set
        """
        self.optSetters[opt](value)

    @Slot(dict)
    def setOptionsFromNode(self, opts: Dict[str, Any]) -> None:
        """Set all options without triggering updates back to the node."""
        for opt, val in opts.items():
            self.setOptionFromNode(opt, val)

    @emitGuiUpdate('optionToNode')
    def signalOption(self, name: str) -> Tuple[str, Any]:
        """Returns name and value of an option.

        Value is determined from the optGetters.
        Decorated with ``@emitGuiUpdate('optionToNode')``.

        :param name: name of the option
        """
        return name, self.optGetters[name]()

    @emitGuiUpdate('allOptionsToNode')
    def signalAllOptions(self) -> Dict[str, Any]:
        """Return all options as a dictionary

        Decorated with ``@emitGuiUpdate('optionToNode')``.
        """
        return self.getAllOptions()



from typing import List

import pyqtgraph as pg

from plottr import config_entry as getcfg


__all__: List[str] = []


bg = getcfg('main', 'pyqtgraph', 'background', default='w')
pg.setConfigOption('background', bg)

fg = getcfg('main', 'pyqtgraph', 'foreground', default='k')
pg.setConfigOption('foreground', fg)



"""
``plottr.plot.mpl.widgets`` -- This module contains general matplotlib plotting tools.
"""

import io
from typing import Tuple, Optional, List, Dict

from numpy import rint
from matplotlib import rcParams
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FCanvas,
    NavigationToolbar2QT as NavBar,
)

from plottr import QtWidgets, QtGui, QtCore, config as plottrconfig
from plottr.data.datadict import DataDictBase
from plottr.gui.tools import widgetDialog, dpiScalingFactor
from ..base import PlotWidget, PlotWidgetContainer


class MPLPlot(FCanvas):
    """
    This is the basic matplotlib canvas widget we are using for matplotlib
    plots. This canvas only provides a few convenience tools for automatic
    sizing, but is otherwise not very different from the class ``FCanvas``
    that comes with matplotlib (and which we inherit).
    It can be used as any QT widget.
    """

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None,
                 width: float = 4.0, height: float = 3.0, dpi: int = 150,
                 constrainedLayout: bool = True):
        """
        Create the canvas.

        :param parent: the parent widget
        :param width: canvas width (inches)
        :param height: canvas height (inches)
        :param dpi: figure dpi
        :param constrainedLayout:
        """

        self.fig = Figure(figsize=(width, height), dpi=dpi,
                          constrained_layout=constrainedLayout)
        super().__init__(self.fig)

        self.axes: List[Axes] = []
        self._tightLayout = False
        self._showInfo = False
        self._infoArtist = None
        self._info = ''
        self._meta_info: Dict[str, str] = {}
        self._constrainedLayout = constrainedLayout

        self.clearFig()
        self.setParent(parent)
        self.setRcParams()

    def autosize(self) -> None:
        """Sets some default spacings/margins."""
        if not self._constrainedLayout:
            self.fig.subplots_adjust(left=0.125, bottom=0.125,
                                     top=0.9, right=0.875,
                                     wspace=0.35, hspace=0.2)
        self.draw()

    def clearFig(self) -> None:
        """clear and reset the canvas."""
        self.fig.clear()
        self.autosize()

    def setRcParams(self) -> None:
        """apply matplotlibrc config from plottr configuration files."""
        cfg = plottrconfig().get('main', {}).get('matplotlibrc', {})
        for k, v in cfg.items():
            rcParams[k] = v
        rcParams['font.size'] = cfg.get('font.size', 6) * dpiScalingFactor(self)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        """
        Re-implementation of the widget resizeEvent method.
        Makes sure we resize the plots appropriately.
        """
        self.autosize()
        super().resizeEvent(event)

    def setShowInfo(self, show: bool) -> None:
        """Whether to show additional info in the plot"""
        self._showInfo = show
        self.updateInfo()

    def updateInfo(self) -> None:
        if self._infoArtist is not None:
            self._infoArtist.remove()
            self._infoArtist = None

        if self._showInfo:
            self._infoArtist = self.fig.text(
                0.02, 0.9, self._info,
                fontsize='small',
                verticalalignment='top',
                bbox=dict(boxstyle='round', facecolor='white', alpha=0.67, lw=1, ec='k')
            )
        self.draw()

    def toClipboard(self) -> None:
        """
        Copy the current canvas to the clipboard.
        """
        buf = io.BytesIO()
        self.fig.savefig(buf, dpi=300, facecolor='w', format='png',
                         transparent=True)

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setImage(QtGui.QImage.fromData(buf.getvalue()))
        buf.close()

    def metaToClipboard(self) -> None:
        clipboard = QtWidgets.QApplication.clipboard()
        meta_info_string = "\n".join(f"{k}: {v}"
                                     for k, v in self._meta_info.items())
        clipboard.setText(meta_info_string)

    def setFigureTitle(self, title: str) -> None:
        """Add a title to the figure."""
        self.fig.suptitle(title,
                          horizontalalignment='center',
                          verticalalignment='top',
                          fontsize='small')
        self.draw()

    def setFigureInfo(self, info: str) -> None:
        """Display an info string in the figure"""
        self._info = info
        self.updateInfo()

    def setMetaInfo(self, meta_info: Dict[str, str]) -> None:
        self._meta_info = meta_info


class MPLPlotWidget(PlotWidget):
    """
    Base class for matplotlib-based plot widgets.
    Per default, add a canvas and the matplotlib NavBar.
    """

    def __init__(self, parent: Optional[PlotWidgetContainer] = None):
        super().__init__(parent=parent)

        #: the plot widget
        self.plot = MPLPlot()

        #: the matplotlib toolbar
        self.mplBar = NavBar(self.plot, self)

        self.addMplBarOptions()
        defaultIconSize = int(16 * dpiScalingFactor(self))
        self.mplBar.setIconSize(QtCore.QSize(defaultIconSize, defaultIconSize))
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.plot)
        layout.addWidget(self.mplBar)
        self.setLayout(layout)

    def setMeta(self, data: DataDictBase) -> None:
        """Add meta info contained in the data to the figure.

        :param data: data object containing the meta information
            if meta field ``title`` or ``info`` are in the data object, then
            they will be added as text info to the figure.
        """
        if data.has_meta('title'):
            self.plot.setFigureTitle(data.meta_val('title'))

        if data.has_meta('info'):
            self.plot.setFigureInfo(data.meta_val('info'))

        all_meta = {}
        for k, v in sorted(data.meta_items()):
            this_meta = str(v)
            if len(this_meta) > 200:
                this_meta = this_meta[:200] + "..."
            all_meta[k] = this_meta
        self.plot.setMetaInfo(all_meta)

    def addMplBarOptions(self) -> None:
        """Add options for displaying ``info`` meta data and copying the figure to the clipboard to the
        plot toolbar."""
        self.mplBar.addSeparator()
        infoAction = self.mplBar.addAction('Show Info')
        infoAction.setCheckable(True)
        infoAction.toggled.connect(self.plot.setShowInfo)

        self.mplBar.addSeparator()
        self.mplBar.addAction('Copy Figure', self.plot.toClipboard)
        self.mplBar.addAction('Copy Meta', self.plot.metaToClipboard)


def figureDialog() -> Tuple[Figure, QtWidgets.QDialog]:
    """Make a dialog window containing a :class:`.MPLPlotWidget`.

    :return: The figure object of the plot, and the dialog window object.
    """
    widget = MPLPlotWidget()
    return widget.plot.fig, widgetDialog(widget)




"""
log.py

Handler and widget for logging in plottr.

Use the setupLogging method to create a logging widget/dialog.

If a logger is accessed via loggin.getLogger('plottr.*') the logging
widget will capture the log and display it.
"""

# TODO: unify with the one from instrumentserver. should maybe go into labcore?

import sys
from typing import Optional, Union
from plottr import QtWidgets, QtGui
import logging

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'

COLORS = {
    logging.ERROR : QtGui.QColor('red'),
    logging.WARNING : QtGui.QColor('orange'),
    logging.INFO : QtGui.QColor('green'),
    logging.DEBUG : QtGui.QColor('gray'),
    }

LEVEL = logging.INFO

class QLogHandler(logging.Handler):

    def __init__(self, parent: QtWidgets.QWidget):
        super().__init__()
        self.widget = QtWidgets.QTextEdit(parent)
        self.widget.setReadOnly(True)

    def emit(self, record: logging.LogRecord) -> None:
        msg = self.format(record)
        clr = COLORS.get(record.levelno, QtGui.QColor('black'))
        self.widget.setTextColor(clr)
        self.widget.append(msg)
        self.widget.verticalScrollBar().setValue(
            self.widget.verticalScrollBar().maximum()
        )


class LogWidget(QtWidgets.QWidget):
    """
    A simple logger widget. Uses QLogHandler as handler.
    Does not do much else.
    """
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None,
                 level: int = logging.INFO):
        super().__init__(parent)

        ### set up the graphical handler
        fmt = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s\n" +
                "    %(message)s",
            datefmt='%Y-%m-%d %H:%M:%S',
            )
        logTextBox = QLogHandler(self)
        logTextBox.setFormatter(fmt)

        # make the widget
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(logTextBox.widget)
        self.setLayout(layout)

        # configure the logger. delete pre-existing graphical handler.
        self.logger = getLogger()
        for h in self.logger.handlers:
            if isinstance(h, QLogHandler):
                self.logger.removeHandler(h)
                h.widget.deleteLater()
                del h

        self.logger.addHandler(logTextBox)
        self.logger.setLevel(level)


    def setLevel(self, level: int) -> None:
        self.logger.setLevel(level)


def logDialog(widget: QtWidgets.QWidget) -> QtWidgets.QDialog:
    layout = QtWidgets.QVBoxLayout()
    d = QtWidgets.QDialog()
    d.setLayout(layout)
    layout.addWidget(widget)
    d.setWindowTitle('Plottr | Log')
    return d


def setupLogging(level: int = logging.INFO,
                 makeDialog: bool = True) -> Union[QtWidgets.QDialog, LogWidget]:
    """
    Setup logging for plottr. Creates the widget and handler.
    if makeDialog is True, embed the widget into the dialog.
    Returns either the widget or the dialog.
    """
    w = LogWidget(level=level)
    if makeDialog:
        d = logDialog(w)
        d.show()
        return d
    else:
        return w


def getLogger(module: str = '') -> logging.Logger:
    """
    Return the logger we use within the plottr framework.
    """
    mod = 'plottr'
    if module != '':
        if module.split('.')[0] == 'plottr':
            mod = module
        else:
            mod += f'.{module}'

    logger = logging.getLogger(mod)
    logger.setLevel(LEVEL)
    return logger


def enableStreamHandler(enable: bool = False) -> None:
    """
    enable/disable output to stderr. Enabling is useful when not
    using the UI logging window.
    """
    logger = getLogger()
    hasStreamHandler = False
    for h in logger.handlers:
        if isinstance(h, logging.StreamHandler):
            hasStreamHandler = True
            if not enable:
                logger.removeHandler(h)
                del h

    if enable and not hasStreamHandler:
        streamHandler = logging.StreamHandler(sys.stderr)
        fmt = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s\n" +
                "    %(message)s",
            datefmt='%Y-%m-%d %H:%M:%S',
            )
        streamHandler.setFormatter(fmt)
        logger.addHandler(streamHandler)



"""tools.py

tools for working with flowcharts and nodes.
"""
from typing import Type, Tuple

from plottr import Flowchart
from plottr.node import Node

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'


def flowchart() -> Flowchart:
    """Create a blank flowchart instance

    :return: new Flowchart, with terminals ``dataIn`` and
             ``dataOut``.
    """
    fc = Flowchart(
        terminals=dict(
            dataIn=dict(io='in'),
            dataOut=dict(io='out'),
        )
    )
    return fc


def linearFlowchart(*nodes: Tuple[str, Type[Node]]) \
        -> Flowchart:
    """Create a flowchart with linearly connected nodes

    :param node: Nodes, in the order in which they will be connected.
                 given in format (name, NodeClass). Nodes will be instantiated
                 in this function. If no nodes are given, the input of the
                 flowchart is directly connected to the output.
    :return: Flowchart with the specified connections made.
    """
    fc = flowchart()

    nNodes = len(nodes)
    instances = []

    if nNodes == 0:
        fc.connectTerminals(fc['dataIn'], fc['dataOut'])
        return fc

    for i, (name, cls) in enumerate(nodes):
        node = cls(name=name)
        instances.append(node)
        fc.addNode(node, name=name)

        if i == 0:
            fc.connectTerminals(fc['dataIn'], node['dataIn'])
        else:
            fc.connectTerminals(instances[i - 1]['dataOut'], node['dataIn'])

        if i == nNodes - 1:
            fc.connectTerminals(node['dataOut'], fc['dataOut'])

    return fc



"""
plottr/apps/autoplot.py : tools for simple automatic plotting.
"""

import logging
import os
import time
import argparse
from typing import Union, Tuple, Optional, Type, List, Any, Type
from packaging import version

from .. import QtCore, Flowchart, Signal, Slot, QtWidgets, QtGui
from .. import log as plottrlog
from ..data.datadict import DataDictBase
from ..data.datadict_storage import DDH5Loader
from ..data.qcodes_dataset import QCodesDSLoader
from ..gui import PlotWindow
from ..gui.widgets import MonitorIntervalInput, SnapshotWidget
from ..node.data_selector import DataSelector
from ..node.dim_reducer import XYSelector
from ..node.filter.correct_offset import SubtractAverage
from ..node.scaleunits import ScaleUnits
from ..node.grid import DataGridder, GridOption
from ..node.tools import linearFlowchart
from ..node.node import Node
from ..node.histogram import Histogrammer
from ..plot import PlotNode, makeFlowchartWithPlot, PlotWidget
from ..plot.pyqtgraph.autoplot import AutoPlot as PGAutoPlot
from ..utils.misc import unwrap_optional

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'


# TODO: * separate logging window


def logger() -> logging.Logger:
    logger = logging.getLogger('plottr.apps.autoplot')
    logger.setLevel(plottrlog.LEVEL)
    return logger


def autoplot(inputData: Union[None, DataDictBase] = None,
             plotWidgetClass: Optional[Type[PlotWidget]] = None) \
        -> Tuple[Flowchart, 'AutoPlotMainWindow']:
    """
    Sets up a simple flowchart consisting of a data selector, gridder,
    an xy-axes selector, and creates a GUI together with an autoplot
    widget.

    :returns: the flowchart object and the dialog widget
    """

    nodes: List[Tuple[str, Type[Node]]] = [
        ('Data selection', DataSelector),
        ('Grid', DataGridder),
        ('Dimension assignment', XYSelector),
    ]

    widgetOptions = {
        "Data selection": dict(visible=True,
                               dockArea=QtCore.Qt.TopDockWidgetArea),
        "Dimension assignment": dict(visible=True,
                                     dockArea=QtCore.Qt.TopDockWidgetArea),
    }

    fc = makeFlowchartWithPlot(nodes)
    win = AutoPlotMainWindow(fc, widgetOptions=widgetOptions,
                             plotWidgetClass=plotWidgetClass)
    win.show()

    if inputData is not None:
        win.setInput(data=inputData)

    return fc, win


class UpdateToolBar(QtWidgets.QToolBar):
    """
    A very simple toolbar to enable monitoring or triggering based on a timer.
    Contains a timer whose interval can be set.
    The toolbar will then emit a signal each interval.
    """

    #: Signal emitted after each trigger interval
    trigger = Signal()

    def __init__(self, name: str, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(name, parent)

        self.monitorInput = MonitorIntervalInput()
        self.monitorInput.setToolTip('Set to 0 for disabling triggering')
        self.monitorInput.intervalChanged.connect(self.setMonitorInterval)
        self.addWidget(self.monitorInput)

        self.monitor = QtCore.QTimer()
        self.monitor.timeout.connect(self.monitorTriggered)

    @Slot()
    def monitorTriggered(self) -> None:
        """
        Is called whenever the monitor timer triggers, and emit the
        :attr:`trigger` Signal.
        """
        logger().debug('Emit trigger')
        self.trigger.emit()

    @Slot(float)
    def setMonitorInterval(self, val: float) -> None:
        """
        Start a background timer that is triggered every `val' seconds.

        :param val: trigger interval in seconds
        """
        self.monitor.stop()
        if val > 0:
            self.monitor.start(int(val * 1000))

        self.monitorInput.spin.setValue(val)

    @Slot()
    def stop(self) -> None:
        """
        Stop the timer.
        """
        self.monitor.stop()


class AutoPlotMainWindow(PlotWindow):

    def __init__(self, fc: Flowchart,
                 parent: Optional[QtWidgets.QMainWindow] = None,
                 monitor: bool = False,
                 monitorInterval: Union[float, None] = None,
                 loaderName: Optional[str] = None,
                 plotWidgetClass: Optional[Type[PlotWidget]] = None,
                 **kwargs: Any):

        super().__init__(parent, fc=fc, plotWidgetClass=plotWidgetClass,
                         **kwargs)

        self.fc = fc
        self.loaderNode: Optional[Node] = None
        if loaderName is not None:
            self.loaderNode = fc.nodes()[loaderName]

        # a flag we use to set reasonable defaults when the first data
        # is processed
        self._initialized = False

        windowTitle = "Plottr | Autoplot"
        self.setWindowTitle(windowTitle)

        # status bar
        self.status = QtWidgets.QStatusBar()
        self.setStatusBar(self.status)

        # menu bar
        self.menu = self.menuBar()
        self.fileMenu = self.menu.addMenu('&Data')

        if self.loaderNode is not None:
            refreshAction = QtWidgets.QAction('&Refresh', self)
            refreshAction.setShortcut('R')
            refreshAction.triggered.connect(self.refreshData)
            self.fileMenu.addAction(refreshAction)

        # add monitor if needed
        if monitor:
            self.monitorToolBar: Optional[UpdateToolBar] = UpdateToolBar('Monitor data')
            self.addToolBar(self.monitorToolBar)
            self.monitorToolBar.trigger.connect(self.refreshData)
            if monitorInterval is not None:
                self.setMonitorInterval(monitorInterval)
        else:
            self.monitorToolBar = None

        # set some sane defaults any time the data is significantly altered.
        if self.loaderNode is not None:
            self.loaderNode.dataFieldsChanged.connect(self.onChangedLoaderData)

    def setMonitorInterval(self, val: float) -> None:
        if self.monitorToolBar is not None:
            self.monitorToolBar.setMonitorInterval(val)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """
        When closing the inspectr window, do some house keeping:
        * stop the monitor, if running
        """
        if self.monitorToolBar is not None:
            self.monitorToolBar.stop()
        return super().closeEvent(event)

    def showTime(self) -> None:
        """
        Displays current time and DS info in the status bar.
        """
        tstamp = time.strftime("%Y-%m-%d %H:%M:%S")
        self.status.showMessage(f"loaded: {tstamp}")

    @Slot()
    def onChangedLoaderData(self) -> None:
        assert self.loaderNode is not None
        data = self.loaderNode.outputValues()['dataOut']
        if data is not None:
            self.setDefaults(self.loaderNode.outputValues()['dataOut'])

    @Slot()
    def refreshData(self) -> None:
        """
        Refresh the dataset by calling `update' on the dataset loader node.
        """
        if self.loaderNode is not None:

            self.loaderNode.update()
            self.showTime()

            if not self._initialized and self.loaderNode.nLoadedRecords > 0:
                self.onChangedLoaderData()
                self._initialized = True

    def setInput(self, data: DataDictBase, resetDefaults: bool = True) -> None:
        """
        Set input to the flowchart. Can only be used when no loader node is
        defined.
        """
        if self.loaderNode is not None:
            logger().warning("A loader node is defined. Use that for inserting data.")
        else:
            self.fc.setInput(dataIn=data)
            if resetDefaults or not self._initialized:
                self.setDefaults(data)
                self._initialized = True

    def setDefaults(self, data: DataDictBase) -> None:
        """
        try to set some reasonable defaults so there's a plot right away.
        """
        selected = data.dependents()
        if len(selected) > 0:
            selected = selected[:1]

        axes = data.axes(selected)
        drs = dict()
        if len(axes) >= 2:
            drs = {axes[-1]: 'x-axis', axes[-2]: 'y-axis'}
        if len(axes) == 1:
            drs = {axes[0]: 'x-axis'}

        try:
            self.fc.nodes()['Data selection'].selectedData = selected
            self.fc.nodes()['Grid'].grid = GridOption.guessShape, {}
            self.fc.nodes()['Dimension assignment'].dimensionRoles = drs
        # FIXME: this is maybe a bit excessive, but trying to set all the defaults
        #   like this can result in many types of errors.
        #   a better approach would be to inspect the data better and make sure
        #   we can set defaults reliably.
        except:
            pass
        unwrap_optional(self.plotWidget).update()


class QCAutoPlotMainWindow(AutoPlotMainWindow):
    """
    Main Window for autoplotting a qcodes dataset.

    Comes with menu options for refreshing the loaded dataset,
    and a toolbar for enabling live-monitoring/refreshing the loaded
    dataset.
    """

    def __init__(self, fc: Flowchart,
                 parent: Optional[QtWidgets.QMainWindow] = None,
                 pathAndId: Optional[Tuple[str, int]] = None, **kw: Any):

        super().__init__(fc, parent, **kw)

        windowTitle = "Plottr | QCoDeS autoplot"
        if pathAndId is not None:
            path = os.path.abspath(pathAndId[0])
            windowTitle += f" | {os.path.split(path)[1]} [{pathAndId[1]}]"
            pathAndId = path, pathAndId[1]
        self.setWindowTitle(windowTitle)

        if pathAndId is not None and self.loaderNode is not None:
            self.loaderNode.pathAndId = pathAndId

        if self.loaderNode is not None and self.loaderNode.nLoadedRecords > 0:
            self.setDefaults(self.loaderNode.outputValues()['dataOut'])
            self._initialized = True

    def setDefaults(self, data: DataDictBase) -> None:
        super().setDefaults(data)
        import qcodes as qc
        qcodes_support = (version.parse(qc.__version__) >=
                          version.parse("0.20.0"))
        if data.meta_val('qcodes_shape') is not None and qcodes_support:
            self.fc.nodes()['Grid'].grid = GridOption.metadataShape, {}
        else:
            self.fc.nodes()['Grid'].grid = GridOption.guessShape, {}


def autoplotQcodesDataset(log: bool = False,
                          pathAndId: Union[Tuple[str, int], None] = None) \
        -> Tuple[Flowchart, QCAutoPlotMainWindow]:
    """
    Sets up a simple flowchart consisting of a data selector,
    an xy-axes selector, and creates a GUI together with an autoplot
    widget.

    returns the flowchart object and the mainwindow widget
    """

    fc = linearFlowchart(
        ('Data loader', QCodesDSLoader),
        ('Data selection', DataSelector),
        ('Grid', DataGridder),
        ('Dimension assignment', XYSelector),
        ('Subtract average', SubtractAverage),
        ('Scale units', ScaleUnits),
        ('plot', PlotNode)
    )

    widgetOptions = {
        "Data selection": dict(visible=True,
                               dockArea=QtCore.Qt.TopDockWidgetArea),
        "Dimension assignment": dict(visible=True,
                                     dockArea=QtCore.Qt.TopDockWidgetArea),
    }

    win = QCAutoPlotMainWindow(fc, pathAndId=pathAndId,
                               widgetOptions=widgetOptions,
                               monitor=True,
                               loaderName='Data loader')
    win.show()

    return fc, win


def autoplotDDH5(filepath: str = '', groupname: str = 'data') \
        -> Tuple[Flowchart, AutoPlotMainWindow]:

    fc = linearFlowchart(
        ('Data loader', DDH5Loader),
        ('Data selection', DataSelector),
        ('Grid', DataGridder),
        ('Histogram', Histogrammer),
        ('Dimension assignment', XYSelector),
        ('plot', PlotNode)
    )

    widgetOptions = {
        "Data selection": dict(visible=True,
                               dockArea=QtCore.Qt.TopDockWidgetArea),
        "Histogram": dict(visible=False,
                          dockArea=QtCore.Qt.TopDockWidgetArea),
        "Dimension assignment": dict(visible=True,
                                     dockArea=QtCore.Qt.TopDockWidgetArea),
    }

    win = AutoPlotMainWindow(fc, loaderName='Data loader',
                             widgetOptions=widgetOptions,
                             monitor=True,
                             monitorInterval=5.0)
    win.show()

    fc.nodes()['Data loader'].filepath = filepath
    fc.nodes()['Data loader'].groupname = groupname
    win.refreshData()
    win.setMonitorInterval(5.0)

    return fc, win


def main(f: str, g: str) -> int:
    app = QtWidgets.QApplication([])
    fc, win = autoplotDDH5(f, g)

    return app.exec_()


def script() -> None:
    parser = argparse.ArgumentParser(
        description='plottr autoplot .dd.h5 files.'
    )
    parser.add_argument('--filepath', help='path to .dd.h5 file',
                        default='')
    parser.add_argument('--groupname', help='group in the hdf5 file',
                        default='data')
    args = parser.parse_args()

    main(args.filepath, args.groupname)



from typing import TYPE_CHECKING, List, Tuple, Dict, Any, Optional
from importlib.abc import Loader
from importlib.util import spec_from_file_location, module_from_spec
import logging
import os
import sys

if TYPE_CHECKING:
    from PyQt5 import QtCore, QtGui, QtWidgets
    Signal = QtCore.pyqtSignal
    Slot = QtCore.pyqtSlot
else:
    from qtpy import QtCore, QtGui, QtWidgets
    Signal = QtCore.Signal
    Slot = QtCore.Slot

from pyqtgraph.flowchart import Flowchart as pgFlowchart, Node as pgNode
Flowchart = pgFlowchart
NodeBase = pgNode

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

logger = logging.getLogger(__name__)
logger.info(f"Imported plottr version: {__version__}")


plottrPath = os.path.split(os.path.abspath(__file__))[0]


def configPaths() -> Tuple[str, str, str]:
    """Get the folders where plottr looks for config files.

    :return: List of absolute paths, in order of priority:
        (1) current working directory
        (2) ~/.plottr
        (3) config directory in the package.
    """
    builtIn = os.path.join(plottrPath, 'config')
    user = os.path.join(os.path.expanduser("~"), '.plottr')
    cwd = os.getcwd()
    return cwd, user, builtIn


def configFiles(fileName: str) -> List[str]:
    """Get available config files with the given file name.

    :param fileName: file name, without path
    :return: List of found config files with the provided name, in order
        or priority.
    """
    ret = []
    for path in configPaths():
        fp = os.path.join(path, fileName)
        if os.path.exists(fp):
            ret.append(fp)
    return ret


def config(names: Optional[List[str]] = None) -> \
        Dict[str, Any]:
    """Return the plottr configuration as a dictionary.

    Each config file found is expected to contain a dictionary with name
    ``config``. The returned configuration is of the form
    ``
    {
        cfg_1: {...},
        cfg_2: {...},
    }
    ``
    The keys in the returned dictionary are the names given, and the contents
    of each entry the dictionary found in the corresponding files.

    Values returned are determined in hierarchical order:
    If configs are found on package and user levels, we first look at the
    package-provided config, and then update with user-provided ones (see doc
    of :func:`.configPaths`). I.e., user-provided config has the highest
    priority and overrides package-provided config.

    Note: currently, exceptions raised when trying to import config objects are
    not caught. Erroneous config files may thus crash the program.

    :param names: List of files. For given ``name`` will look
        for ``plottrcfg_<name>.py`` in the config directories.
        if ``None``, will look only for ``plottrcfg_main.py``
    :param forceReload: If True, will not use a cached config file if present.
        will thus get the most recent config from file, without need to restart
        the program.
    """
    if names is None:
        names = ['main']

    config = {}
    for name in names:
        modn = f"plottrcfg_{name}"
        filen = f"{modn}.py"
        this_cfg: Dict[str, Any] = {}
        for filep in configFiles(filen)[::-1]:
            spec = spec_from_file_location(modn, filep)
            if spec is None:
                raise FileNotFoundError(f"Could not locate spec for {modn}, {filep}")
            mod = module_from_spec(spec)
            sys.modules[modn] = mod
            assert isinstance(spec.loader, Loader)
            spec.loader.exec_module(mod)
            this_cfg.update(getattr(mod, 'config', {}))

        config[name] = this_cfg
    return config


def config_entry(*path: str, default: Optional[Any] = None,
                 names: Optional[List[str]] = None) -> Any:
    """Get a specific config value.

    ..Example: If the config is:: python

        config = {
            'foo' : {
                'bar' : 'spam',
            },
        }

    .. then we can get an entry like this:: python

        >>> config_entry('foo', 'bar', default=None)
        'spam'
        >>> config_entry('foo', 'bacon')
        None
        >>> config_entry('foo', 'bar', 'bacon')
        None

    :param path: strings denoting the nested keys to the desired value
    :param names: see :func:`.config`.
    :param default: what to return when key isn't found in the config.
    :returns: desired value
    """

    cfg: Any = config(names)
    for k in path:
        if isinstance(cfg, dict) and k in cfg:
            cfg = cfg.get(k)
        else:
            return default
    return cfg



from matplotlib import cycler
from plottr.plot.pyqtgraph.autoplot import AutoPlot as PGAutoPlot
from plottr.plot.mpl.autoplot import AutoPlot as MPLAutoPlot

config = {

    'default-plotwidget': MPLAutoPlot,

    'matplotlibrc': {
        'axes.grid': True,
        'axes.prop_cycle': cycler('color', ['1f77b4', 'ff7f0e', '2ca02c', 'd62728', '9467bd', '8c564b',
                                            'e377c2', '7f7f7f', 'bcbd22', '17becf']),
        'figure.dpi': 150,
        'figure.figsize': (4.5, 3),
        'font.size': 6,
        'font.family': ['Helvetica', 'Arial', 'DejaVu Sans', 'Bitstream Vera Sans'],
        'grid.linewidth': 0.5,
        'grid.linestyle': '--',
        'image.cmap': 'magma',
        'legend.fontsize': 6,
        'legend.frameon': True,
        'legend.numpoints': 1,
        'legend.scatterpoints': 1,
        'lines.marker': 'o',
        'lines.markersize': '3',
        'lines.markeredgewidth': 1,
        'lines.markerfacecolor': 'w',
        'lines.linestyle': '-',
        'lines.linewidth': 1,
        'savefig.dpi': 300,
        'savefig.transparent': False,
    },

    'pyqtgraph': {
        'background': 'w',
        'foreground': 'k',
        'line_colors': ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
                        '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'],
        'line_symbols': ['o', ],
        'line_symbol_size': 7,
        'minimum_plot_size': (400, 400),
    }
}



from typing import Dict, Any, Callable, Optional, Type

from .. import QtGui, QtWidgets
from .node import Node, NodeWidget, updateOption

connectCallableType = Callable[['AutoNodeGuiTemplate', str, Dict[str, Any], bool], QtWidgets.QWidget]


class AutoNodeGuiTemplate(NodeWidget):
    widgetConnection: Dict[type, connectCallableType] = dict()


def connectIntegerSpinbox(
        gui: AutoNodeGuiTemplate,
        optionName: str,
        specs: Dict[str, Any],
        confirm: bool) -> QtWidgets.QSpinBox:

    widget = QtWidgets.QSpinBox()
    widget.setValue(specs.get('initialValue', 1))
    if not confirm:
        widget.valueChanged.connect(lambda x: gui.signalOption(optionName))

    gui.optGetters[optionName] = widget.value
    gui.optSetters[optionName] = widget.setValue

    return widget

def connectFloatSpinbox(
        gui: AutoNodeGuiTemplate,
        optionName: str,
        specs: Dict[str, Any],
        confirm: bool) -> QtWidgets.QDoubleSpinBox:

    widget = QtWidgets.QDoubleSpinBox()
    widget.setValue(specs.get('initialValue', 1))
    if not confirm:
        widget.valueChanged.connect(lambda x: gui.signalOption(optionName))

    gui.optGetters[optionName] = widget.value
    gui.optSetters[optionName] = widget.setValue

    return widget


class AutoNodeGui(AutoNodeGuiTemplate):

    widgetConnection = {
        int: connectIntegerSpinbox,
        float: connectFloatSpinbox,
    }

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None,
                 node: Optional[Node] = None):
        super().__init__(parent)
        layout = QtWidgets.QFormLayout()
        self.setLayout(layout)

    def addOption(self, name: str, specs: Dict[str, Any], confirm: bool) -> None:
        optionType = specs.get('type', None)
        widget = None
        func = self.widgetConnection.get(optionType, None)
        if func is not None:
            widget = func(self, name, specs, confirm)

        self.layout().addRow(name, widget)

    def addConfirm(self) -> None:
        widget = QtWidgets.QPushButton('Confirm')
        widget.pressed.connect(self.signalAllOptions)
        self.layout().addRow('', widget)


class AutoNode(Node):

    def addOption(self, name: str, specs: Dict[str, Any]) -> None:
        varname = '_' + name
        setattr(self, varname, specs.get("initialValue", None))

        def getter(self: AutoNode) -> Any:
            return getattr(self, varname)

        @updateOption(name)
        def setter(self: AutoNode, val: str) -> None:
            setattr(self, varname, val)

        setattr(self.__class__, name, property(getter, setter))


def autonode(nodeName: str, confirm: bool = True, **options: Any) -> Callable:

    def decorator(func: Callable) -> Type["AutoNode"]:

        class AutoNodeGui_(AutoNodeGui):
            def __init__(self, parent: Optional[QtWidgets.QWidget] = None,
                         node: Optional[Node] = None):
                super().__init__(parent)
                for optName, optSpecs in options.items():
                    self.addOption(optName, optSpecs, confirm=confirm)
                if confirm:
                    self.addConfirm()

        class AutoNode_(AutoNode):
            def __init__(self, name: str):
                super().__init__(name)
                for optName, optSpecs in options.items():
                    self.addOption(optName, optSpecs)

        AutoNode_.__name__ = nodeName
        AutoNode_.nodeName = nodeName
        AutoNode_.nodeOptions = options
        AutoNode_.process = func  # type: ignore[assignment]
        AutoNode_.useUi = True
        AutoNode_.uiClass = AutoNodeGui_

        return AutoNode_

    return decorator



"""
qcodes_dataset.py

Dealing with qcodes dataset (the database) data in plottr.
"""
import os
from itertools import chain
from operator import attrgetter
from typing import Dict, List, Set, Union, TYPE_CHECKING, Any, Tuple, Optional, cast

from typing_extensions import TypedDict

import pandas as pd

from qcodes.dataset.data_set import load_by_id
from qcodes.dataset.experiment_container import experiments
from qcodes.dataset.sqlite.database import initialise_or_create_database_at

from .datadict import DataDictBase, DataDict, combine_datadicts
from ..node.node import Node, updateOption

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'

if TYPE_CHECKING:
    try:
        from qcodes.dataset import DataSetProtocol
    except ImportError:
        from qcodes.dataset.data_set import DataSet as DataSetProtocol
    from qcodes import ParamSpec


def _get_names_of_standalone_parameters(paramspecs: List['ParamSpec']
                                        ) -> Set[str]:
    all_independents = set(spec.name
                           for spec in paramspecs
                           if len(spec.depends_on_) == 0)
    used_independents = set(d for spec in paramspecs for d in spec.depends_on_)
    standalones = all_independents.difference(used_independents)
    return standalones


class IndependentParameterDict(TypedDict):
    unit: str
    label: str
    values: List[Any]


class DependentParameterDict(IndependentParameterDict):
    axes: List[str]


DataSetStructureDict = Dict[str, Union[IndependentParameterDict, DependentParameterDict]]


class DataSetInfoDict(TypedDict):
    experiment: str
    sample: str
    name: str
    completed_date: str
    completed_time: str
    started_date: str
    started_time: str
    structure: Optional[DataSetStructureDict]
    records: int
    guid: str
    inspectr_tag: str


# Tools for extracting information on runs in a database


def get_ds_structure(
        ds: 'DataSetProtocol'
) -> DataSetStructureDict:
    """
    Return the structure of the dataset, i.e., a dictionary in the form
        {
            'dependent_parameter_name': {
                'unit': unit,
                'label': label,
                'axes': list of names of independent parameters,
                'values': []
            },
            'independent_parameter_name': {
                'unit': unit,
                'label': label,
                'values': []
            },
            ...
        }

    Note that standalone parameters (those which don't depend on any other
    parameter and no other parameter depends on them) are not included
    in the returned structure.
    """

    structure: DataSetStructureDict = {}

    paramspecs = ds.get_parameters()

    standalones = _get_names_of_standalone_parameters(paramspecs)

    for spec in paramspecs:
        if spec.name not in standalones:
            if len(spec.depends_on_) > 0:
                structure[spec.name] = DependentParameterDict(unit=spec.unit,
                                                              label=spec.label,
                                                              values=[],
                                                              axes=list(spec.depends_on_))
            else:
                structure[spec.name] = IndependentParameterDict(unit=spec.unit,
                                                                label=spec.label,
                                                                values=[])
    return structure


def get_ds_info(ds: 'DataSetProtocol', get_structure: bool = True) -> DataSetInfoDict:
    """
    Get some info on a DataSet in dict.

    if get_structure is True: return the datastructure in that dataset
    as well (key is `structure' then).
    """
    _complete_ts = ds.completed_timestamp()
    if _complete_ts is not None:
        completed_date = _complete_ts[:10]
        completed_time = _complete_ts[11:]
    else:
        completed_date = ''
        completed_time = ''

    _start_ts = ds.run_timestamp()
    if _start_ts is not None:
        started_date = _start_ts[:10]
        started_time = _start_ts[11:]
    else:
        started_date = ''
        started_time = ''

    if get_structure:
        structure: Optional[DataSetStructureDict] = get_ds_structure(ds)
    else:
        structure = None

    data = DataSetInfoDict(
        experiment=ds.exp_name,
        sample=ds.sample_name,
        name=ds.name,
        completed_date=completed_date,
        completed_time=completed_time,
        started_date=started_date,
        started_time=started_time,
        structure=structure,
        records=ds.number_of_results,
        guid=ds.guid,
        inspectr_tag=ds.metadata.get('inspectr_tag', ''),
    )

    return data


def load_dataset_from(path: str, run_id: int) -> 'DataSetProtocol':
    """
    Loads ``DataSet`` with the given ``run_id`` from a database file that
    is located in in the given ``path``.

    Note that after the call to this function, the database location in the
    qcodes config of the current python process is changed to ``path``.
    """
    initialise_or_create_database_at(path)
    return load_by_id(run_id=run_id)


def get_runs_from_db(path: str, start: int = 0,
                     stop: Union[None, int] = None,
                     get_structure: bool = False) -> Dict[int, DataSetInfoDict]:
    """
    Get a db ``overview`` dictionary from the db located in ``path``. The
    ``overview`` dictionary maps ``DataSet.run_id``s to dataset information as
    returned by ``get_ds_info`` functions.

    `start` and `stop` refer to indices of the runs in the db that we want
    to have details on; if `stop` is None, we'll use runs until the end.

    If `get_structure` is True, include info on the run data structure
    in the return dict.
    """
    initialise_or_create_database_at(path)

    datasets = sorted(
        chain.from_iterable(exp.data_sets() for exp in experiments()),
        key=attrgetter('run_id')
    )

    # There is no need for checking whether ``stop`` is ``None`` because if
    # it is the following is simply equivalent to ``datasets[start:]``
    datasets = datasets[start:stop]

    overview = {ds.run_id: get_ds_info(ds, get_structure=get_structure)
                for ds in datasets}
    return overview


def get_runs_from_db_as_dataframe(path: str) -> pd.DataFrame:
    """
    Wrapper around `get_runs_from_db` that returns the overview
    as pandas dataframe.
    """
    overview = get_runs_from_db(path)
    df = pd.DataFrame.from_dict(overview, orient='index')
    return df


# Extracting data

def ds_to_datadicts(ds: 'DataSetProtocol') -> Dict[str, DataDict]:
    """
    Make DataDicts from a qcodes DataSet.

    :param ds: qcodes dataset
    :returns: dictionary with one item per dependent.
              key: name of the dependent
              value: DataDict containing that dependent and its
                     axes.
    """
    ret = {}
    has_cache = hasattr(ds, 'cache')
    if has_cache:
        pdata = ds.cache.data()
    else:
        # qcodes < 0.17
        pdata = ds.get_parameter_data()
    for p, spec in ds.paramspecs.items():
        if spec.depends_on != '':
            axes = spec.depends_on_
            data = dict()
            data[p] = dict(unit=spec.unit, label=spec.label, axes=axes, values=pdata[p][p])
            for ax in axes:
                axspec = ds.paramspecs[ax]
                data[ax] = dict(unit=axspec.unit, label=axspec.label, values=pdata[p][ax])
            ret[p] = DataDict(**data)
            ret[p].validate()

    return ret


def ds_to_datadict(ds: 'DataSetProtocol') -> DataDictBase:
    ddicts = ds_to_datadicts(ds)
    ddict = combine_datadicts(*[v for k, v in ddicts.items()])
    return ddict


### qcodes dataset loader node

class QCodesDSLoader(Node):
    nodeName = 'QCodesDSLoader'
    uiClass = None
    useUi = False

    def __init__(self, *arg: Any, **kw: Any):
        self._pathAndId: Tuple[Optional[str], Optional[int]] = (None, None)
        self.nLoadedRecords = 0
        self._dataset: Optional[DataSetProtocol] = None

        super().__init__(*arg, **kw)

    ### Properties

    @property
    def pathAndId(self) -> Tuple[Optional[str], Optional[int]]:
        return self._pathAndId

    # see https://github.com/python/mypy/issues/1362
    @pathAndId.setter  # type: ignore
    @updateOption('pathAndId')
    def pathAndId(self, val: Tuple[Optional[str], Optional[int]]) -> None:
        if val != self.pathAndId:
            self._pathAndId = val
            self.nLoadedRecords = 0
            self._dataset = None

    def process(self, dataIn: Optional[DataDictBase] = None) -> Optional[Dict[str, Any]]:
        if dataIn is not None:
            raise RuntimeError("QCodesDSLoader.process does not take a dataIn argument")
        if None not in self._pathAndId:
            path, runId = cast(Tuple[str, int], self._pathAndId)

            if self._dataset is None:
                self._dataset = load_dataset_from(path, runId)

            if self._dataset.number_of_results > self.nLoadedRecords:

                guid = self._dataset.guid

                experiment_name = self._dataset.exp_name
                sample_name = self._dataset.sample_name
                dataset_name = self._dataset.name

                run_timestamp = self._dataset.run_timestamp()
                completed_timestamp = self._dataset.completed_timestamp()

                title = f"{os.path.split(path)[-1]} | " \
                        f"run ID: {runId} | GUID: {guid}" \
                        "\n" \
                        f"{sample_name} | {experiment_name} | {dataset_name}"

                info = f"""Started: {run_timestamp}
Finished: {completed_timestamp}
DB-File [ID]: {path} [{runId}]"""

                data = ds_to_datadict(self._dataset)

                data.add_meta('qcodes_experiment_name', experiment_name)
                data.add_meta('qcodes_sample_name', sample_name)
                data.add_meta('qcodes_dataset_name', dataset_name)

                data.add_meta('title', title)
                data.add_meta('info', info)

                data.add_meta('qcodes_guid', guid)
                data.add_meta('qcodes_db', path)
                data.add_meta('qcodes_runId', runId)
                data.add_meta('qcodes_completedTS', completed_timestamp)
                data.add_meta('qcodes_runTS', run_timestamp)
                qcodes_shape = getattr(self._dataset.description, "shapes", None)
                data.add_meta('qcodes_shape', qcodes_shape)

                self.nLoadedRecords = self._dataset.number_of_results

                return dict(dataOut=data)

        return None



"""
``plottr.plot.base`` -- Contains the base classes for plotting nodes and widgets.
Everything in here is independent of actual plotting backend, and does not contain plotting commands.
"""

from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum, unique, auto
from types import TracebackType
from typing import Dict, List, Type, Tuple, Optional, Any, \
    OrderedDict as OrderedDictType, Union

import numpy as np

from .. import Signal, Flowchart, QtWidgets
from ..data.datadict import DataDictBase, DataDict, MeshgridDataDict
from ..node import Node, linearFlowchart
from ..utils import LabeledOptions

__author__ = 'Wolfgang Pfaff'
__license__ = 'MIT'


class PlotNode(Node):
    """
    Basic Plot Node, derived from :class:`plottr.node.node.Node`.

    At the moment this doesn't do much besides passing data to the plotting widget.
    Data is just passed through.
    On receipt of new data, :attr:`newPlotData` is emitted.
    """
    nodeName = 'Plot'

    #: Signal emitted when :meth:`process` is called, with the data passed to
    #: it as argument.
    newPlotData = Signal(object)

    def __init__(self, name: str):
        """Constructor for :class:`PlotNode`. """
        super().__init__(name=name)
        self.plotWidgetContainer: Optional['PlotWidgetContainer'] = None

    def setPlotWidgetContainer(self, w: 'PlotWidgetContainer') -> None:
        """Set the plot widget container.

        Makes sure that newly arriving data is sent to plot GUI elements.

        :param w: container to connect the node to.
        """
        self.plotWidgetContainer = w
        self.newPlotData.connect(self.plotWidgetContainer.setData)

    def process(self, dataIn: Optional[DataDictBase] = None) -> Dict[str, Optional[DataDictBase]]:
        """Emits the :attr:`newPlotData` signal when called.
        Note: does not call the parent method :meth:`plottr.node.node.Node.process`.

        :param dataIn: input data
        :returns: input data as is: ``{dataOut: dataIn}``
        """
        self.newPlotData.emit(dataIn)
        return dict(dataOut=dataIn)


class PlotWidgetContainer(QtWidgets.QWidget):
    """
    This is the base widget for Plots, derived from `QWidget`.

    This widget does not implement any plotting. It merely is a wrapping
    widget that contains the actual plot widget in it. This actual plot
    widget can be set dynamically.

    Use :class:`PlotWidget` as base for implementing widgets that can be
    added to this container.
    """

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        """Constructor for :class:`PlotWidgetContainer`. """
        super().__init__(parent=parent)

        self.plotWidget: Optional["PlotWidget"] = None
        self.data: Optional[DataDictBase] = None

        layout: QtWidgets.QVBoxLayout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def setPlotWidget(self, widget: "PlotWidget") -> None:
        """Set the plot widget.

        Makes sure that the added widget receives new data.

        :param widget: plot widget
        """

        # TODO: disconnect everything, make sure old widget is garbage collected

        if widget is self.plotWidget:
            return

        if self.plotWidget is not None:
            self.layout().removeWidget(self.plotWidget)
            self.plotWidget.deleteLater()

        self.plotWidget = widget
        if self.plotWidget is not None:
            self.layout().addWidget(widget)
            self.plotWidget.setData(self.data)

    def setData(self, data: DataDictBase) -> None:
        """set Data. If a plot widget is defined, call the widget's
        :meth:`PlotWidget.setData` method.

        :param data: input data to be plotted.
        """
        self.data = data
        if self.plotWidget is not None:
            self.plotWidget.setData(self.data)


class PlotWidget(QtWidgets.QWidget):
    """
    Base class for Plot Widgets, this just defines the API. Derived from
    `QWidget`.

    Implement a child class for actual plotting.
    """

    def __init__(self, parent: Optional[PlotWidgetContainer] = None) -> None:
        super().__init__(parent=parent)

        self.data: Optional[DataDictBase] = None
        self.dataType: Optional[Type[DataDictBase]] = None
        self.dataStructure: Optional[DataDictBase] = None
        self.dataShapes: Optional[Dict[str, Tuple[int, ...]]] = None
        self.dataLimits: Optional[Dict[str, Tuple[float, float]]] = None
        self.dataChanges: Dict[str, bool] = {
            'dataTypeChanged': False,
            'dataStructureChanged': False,
            'dataShapesChanged': False,
            'dataLimitsChanged': False,
        }

    def updatePlot(self) -> None:
        return None

    def setData(self, data: Optional[DataDictBase]) -> None:
        """Set data. Use this to trigger plotting.

        :param data: data to be plotted.
        """
        self.data = data
        self.dataChanges = self.analyzeData(data)

    def analyzeData(self, data: Optional[DataDictBase]) -> Dict[str, bool]:
        """checks data and compares with previous properties.

        :param data: incoming data to compare to already existing data in the object.
        :return: dictionary with information on what has changed from previous to new data.
            contains key/value pairs where the key is the property analyzed, and the value is True of False. Keys are:

            * `dataTypeChanged` -- has the data class changed?
            * `dataStructureChanged` -- has the internal structure (data fields, etc) changed?
            * `dataShapesChanged` -- have the data fields changed shape?
            * `dataLimitsChanged` -- have the maxima/minima of the data fields changed?

        """
        if data is not None:
            dataType: Optional[Type[DataDictBase]] = type(data)
        else:
            dataType = None

        if data is None:
            dataStructure: Optional[DataDictBase] = None
            dataShapes: Optional[Dict[str, Tuple[int, ...]]] = None
            dataLimits: Optional[Dict[str, Tuple[Any, Any]]] = None
        else:
            dataStructure = data.structure(include_meta=False)
            dataShapes = data.shapes()
            dataLimits = {}
            for n in data.axes() + data.dependents():
                vals = data.data_vals(n)
                dataLimits[n] = vals.min(), vals.max()

        result = {
            'dataTypeChanged': dataType != self.dataType,
            'dataStructureChanged': dataStructure != self.dataStructure,
            'dataShapesChanged': dataShapes != self.dataShapes,
            'dataLimitsChanged': dataLimits != self.dataLimits,
        }

        self.dataType = dataType
        self.dataStructure = dataStructure
        self.dataShapes = dataShapes
        self.dataLimits = dataLimits
        return result

    def dataIsComplex(self, dependentName: Optional[str] = None) -> bool:
        """Determine whether our data is complex.

        :param dependentName: name of the dependent to check. if `None`, check all.
        :return: `True` if data is complex, `False` if not.
        """
        if self.data is None:
            return False

        if dependentName is None:
            for d in self.data.dependents():
                if np.issubsctype(self.data.data_vals(d), np.complexfloating):
                    return True
        else:
            if np.issubsctype(self.data.data_vals(dependentName), np.complexfloating):
                return True

        return False


def makeFlowchartWithPlot(nodes: List[Tuple[str, Type[Node]]],
                          plotNodeName: str = 'plot') -> Flowchart:
    """create a linear FlowChart terminated with a plot node.

    :param nodes: List of Node classes, in the order they are to be arranged.
    :param plotNodeName: name of the plot node that will be appended.
    :return: the resulting FlowChart instance
    """
    nodes.append((plotNodeName, PlotNode))
    fc = linearFlowchart(*nodes)
    return fc


# Types of plots and plottable data
@unique
class PlotDataType(Enum):
    """Types of (plottable) data"""

    #: unplottable data
    unknown = auto()

    #: scatter-type data with 1 dependent (data is not on a grid)
    scatter1d = auto()

    #: line data with 1 dependent (data is on a grid)
    line1d = auto()

    #: scatter data with 2 dependents (data is not on a grid)
    scatter2d = auto()

    #: grid data with 2 dependents
    grid2d = auto()


class ComplexRepresentation(LabeledOptions):
    """Options for plotting complex-valued data."""

    #: only real
    real = "Real"

    #: real and imaginary
    realAndImag = "Real/Imag"

    #: real and imaginary, separated
    realAndImagSeparate = "Real/Imag (split)"

    #: magnitude and phase
    magAndPhase = "Mag/Phase"


def determinePlotDataType(data: Optional[DataDictBase]) -> PlotDataType:
    """
    Analyze input data and determine most likely :class:`PlotDataType`.

    Analysis is simply based on number of dependents and data type.

    :param data: data to analyze.
    :return: type of plot data inferred
    """
    # TODO:
    #   there's probably ways to be more liberal about what can be plotted.
    #   like i can always make a 1d scatter...

    # a few things will result in unplottable data:
    # * wrong data format
    if not isinstance(data, DataDictBase):
        return PlotDataType.unknown

    # * incompatible independents
    if not data.axes_are_compatible():
        return PlotDataType.unknown

    # * too few or too many independents
    if len(data.axes()) < 1 or len(data.axes()) > 2:
        return PlotDataType.unknown

    # * no data to plot
    if len(data.dependents()) == 0:
        return PlotDataType.unknown

    if isinstance(data, MeshgridDataDict):
        shape = data.shapes()[data.dependents()[0]]

        if len(data.axes()) == 2:
            return PlotDataType.grid2d
        else:
            return PlotDataType.line1d

    elif isinstance(data, DataDict):
        if len(data.axes()) == 2:
            return PlotDataType.scatter2d
        else:
            return PlotDataType.scatter1d

    return PlotDataType.unknown


@dataclass
class PlotItem:
    """Data class describing a plot item in :class:`.AutoFigureMaker`."""
    #: List of data arrays (independents and one dependent)
    data: List[Union[np.ndarray, np.ma.MaskedArray]]
    #: unique ID of the plot item
    id: int
    #: ID of the subplot the item will be plotted in
    subPlot: int
    #: type of plot data (unknown is typically OK)
    plotDataType: PlotDataType = PlotDataType.unknown
    #: labels of the data arrays
    labels: Optional[List[str]] = None
    #: options to be passed to plotting functions (depends on backend). Could be formatting options, for example.
    plotOptions: Optional[Dict[str, Any]] = None
    #: return value from the plot command (like matplotlib Artists)
    plotReturn: Optional[Any] = None


@dataclass
class SubPlot:
    """Data class describing a subplot in a :class:`.AutoFigureMaker`."""
    #: ID of the subplot (unique per figure)
    id: int
    #: list of subplot objects (type depends on backend)
    axes: Optional[List[Any]] = None


class AutoFigureMaker:
    """A class for semi-automatic creation of plot figures.
    It must be inherited to tie it to a specific plotting backend.

    The main purpose of this class is to (a) implement actual plotting of
    plot items, and (b) distribute plot items correctly among subpanels of
    a figure.

    FigureMaker is a context manager. The user should eventually only need to
    add data and specify what kind of data it is. FigureMaker will then
    generate plots from that.

    In the simplest form, usage looks something like this::

        >>> with AutoFigureMaker() as fm:
        >>>     fm.addData(x, y, [...])
        >>>     [...]

    See :meth:`addData` for details on how to specify data and how to pass
    plot options to it.
    """

    # TODO: implement feature for always plotting certain traces with other
    #   other ones ('children'). This is mainly used for models/fits.
    #   needs a system to copy certain style aspects from the parents.
    # TODO: similar, but with siblings (imagine Re/Im parts)

    def __init__(self) -> None:

        #: subplots to create
        self.subPlots: OrderedDictType[int, SubPlot] = OrderedDict()

        #: items that will be plotted
        self.plotItems: OrderedDictType[int, PlotItem] = OrderedDict()

        #: ids of all main plot items (does not contain derived/secondary plot items)
        self.plotIds: List = []

        #: ids of all plot items, incl those who are 'joined' with 'main' plot items.
        self.allPlotIds: List = []

        #: how to represent complex data.
        #: must be set before adding data to the plot to have an effect.
        self.complexRepresentation: ComplexRepresentation = ComplexRepresentation.realAndImag

        #: whether to combine 1D traces into one plot
        self.combineTraces: bool = False

    def __enter__(self) -> "AutoFigureMaker":
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:

        self._makeAxes()
        for id in self.subPlots.keys():
            self._makeSubPlot(id)

    # private methods
    def _makeAxes(self) -> None:
        n = self.nSubPlots()
        for id, axes in zip(range(n), self.makeSubPlots(n)):
            if not isinstance(axes, list):
                axes = [axes]
            self.subPlots[id] = SubPlot(id, axes)
        return None

    def _makeSubPlot(self, id: int) -> None:
        items = self.subPlotItems(id)
        for name, item in items.items():
            item.plotReturn = self.plot(item)
        self.formatSubPlot(id)
        return None

    def _splitComplexData(self, plotItem: PlotItem) -> List[PlotItem]:
        if plotItem.labels is None:
            plotItem.labels = [''] * len(plotItem.data)
        label = plotItem.labels[-1]

        if not np.issubsctype(plotItem.data[-1], np.complexfloating):
            return [plotItem]

        elif self.complexRepresentation is ComplexRepresentation.real:
            plotItem.data[-1] = plotItem.data[-1].real
            assert isinstance(plotItem.labels, list)
            if label == '':
                plotItem.labels[-1] = 'Real'
            else:
                plotItem.labels[-1] = label + ' (Real)'
            return [plotItem]

        elif self.complexRepresentation in \
                [ComplexRepresentation.realAndImag, ComplexRepresentation.realAndImagSeparate]:

            re_data = plotItem.data[-1].real
            im_data = plotItem.data[-1].imag

            if label == '':
                re_label, im_label = 'Real', 'Imag'
            else:
                re_label, im_label = label + ' (Real)', label + ' (Imag)'

            re_plotItem = plotItem
            im_plotItem = deepcopy(re_plotItem)

            re_plotItem.data[-1] = re_data
            im_plotItem.data[-1] = im_data
            im_plotItem.id = re_plotItem.id + 1

            if self.complexRepresentation == ComplexRepresentation.realAndImagSeparate \
                    or len(plotItem.data) > 2:
                im_plotItem.subPlot = re_plotItem.subPlot + 1

            # this is a bit of a silly check (see top of the function -- should certainly be True!).
            # but it keeps mypy happy.
            assert isinstance(re_plotItem.labels, list)
            re_plotItem.labels[-1] = re_label
            assert isinstance(im_plotItem.labels, list)
            im_plotItem.labels[-1] = im_label

            return [re_plotItem, im_plotItem]

        else:  # means that self.complexRepresentation is ComplexRepresentation.magAndPhase:
            data = plotItem.data[-1]

            # this check avoids a numpy ComplexWarning when we're working with MaskedArray (almost always)
            mag_data = np.ma.abs(data).real if isinstance(data, np.ma.MaskedArray) else np.abs(data)
            phase_data = np.angle(data)

            if label == '':
                mag_label, phase_label = 'Mag', 'Phase'
            else:
                mag_label, phase_label = label + ' (Mag)', label + ' (Phase)'

            mag_plotItem = plotItem
            phase_plotItem = deepcopy(mag_plotItem)

            mag_plotItem.data[-1] = mag_data
            phase_plotItem.data[-1] = phase_data
            phase_plotItem.id = mag_plotItem.id + 1
            phase_plotItem.subPlot = mag_plotItem.subPlot + 1

            # this is a bit of a silly check (see top of the function -- should certainly be True!).
            # but it keeps mypy happy.
            assert isinstance(mag_plotItem.labels, list)
            mag_plotItem.labels[-1] = mag_label
            assert isinstance(phase_plotItem.labels, list)
            phase_plotItem.labels[-1] = phase_label

            return [mag_plotItem, phase_plotItem]

    # public methods
    def addSubPlot(self) -> int:
        """Add a new subplot.

        :return: ID of the new subplot.
        """
        id = _generate_auto_dict_key(self.subPlots)
        self.subPlots[id] = SubPlot(id, [])
        return id

    def nSubPlots(self) -> int:
        """Count the subplots in the figure.

        :return: number of subplots
        """
        ids = []
        for id, item in self.plotItems.items():
            ids.append(item.subPlot)
        return len(set(ids))

    def subPlotItems(self, subPlotId: int) -> OrderedDictType[int, PlotItem]:
        """Get items in a given subplot.

        :param subPlotId: ID of the subplot
        :return: Dictionary with all plot items and their ids.
        """
        items = OrderedDict()
        for id, item in self.plotItems.items():
            if item.subPlot == subPlotId:
                items[id] = item
        return items

    def subPlotLabels(self, subPlotId: int) -> List[List[str]]:
        """Get the data labels for a given subplot.

        :param subPlotId: ID of the subplot.
        :return: a list with one element per plot item in the subplot.
            Each element contains a list of the labels for that item.
        """
        ret: List[List[str]] = []
        items = self.subPlotItems(subPlotId)
        for id, item in items.items():
            if item.labels is not None:
                for i, l in enumerate(item.labels):
                    while (len(ret)) <= i:
                        ret.append([])
                    ret[i].append(l)
        return ret

    def addData(self, *data: Union[np.ndarray, np.ma.MaskedArray],
                join: Optional[int] = None,
                labels: Optional[List[str]] = None,
                plotDataType: PlotDataType = PlotDataType.unknown,
                **plotOptions: Any) -> int:
        """Add data to the figure.

        :param data: data arrays describing the plot (one or more independents, one dependent)
        :param join: ID of a plot item the new item should be shown together with in the same subplot
        :param labels: list of labels for the data arrays
        :param plotDataType: what kind of plot data the supplied data contains.
        :param plotOptions: options (as kwargs) to be passed to the actual plot functions (depends on the backend)
        :return: ID of the new plot item.
        """

        if self.combineTraces and join is None:
            prev = self.previousPlotId()
            if prev is not None:
                if len(data) == 2 and len(self.plotItems[prev].data) == 2:
                    join = prev

        id = _generate_auto_dict_key(self.plotItems)

        # TODO: allow any negative number
        if join == -1:
            if len(self.plotItems) > 0:
                join = self.previousPlotId()
            else:
                join = None
        if join is None:
            subPlotId = self.nSubPlots()
        else:
            subPlotId = self.plotItems[join].subPlot

        if labels is None:
            labels = [''] * len(data)
        elif len(labels) < len(data):
            labels += [''] * (len(data) - len(labels))

        plotItem = PlotItem(list(data), id, subPlotId,
                            plotDataType, labels, plotOptions)

        for p in self._splitComplexData(plotItem):
            self.plotItems[p.id] = p
            self.allPlotIds.append(p.id)
        self.plotIds.append(id)
        return id

    def previousPlotId(self) -> Optional[int]:
        """Get the ID of the most recently added plot item.
        :return: the ID.
        """
        if not len(self.plotIds) > 0:
            return None

        if len(self.plotIds) > 0:
            return self.plotIds[-1]
        else:
            return None

    def findPlotIndexInSubPlot(self, plotId: int) -> int:
        """find the index of a plot in its subplot

        :param plotId: plot ID to check
        :return: index at which the plot is located in its subplot.
        """
        if plotId not in self.allPlotIds:
            raise ValueError("Plot ID not found.")

        subPlotId = self.plotItems[plotId].subPlot
        itemsInSubPlot = [i for i in self.allPlotIds if self.plotItems[i].subPlot == subPlotId]
        return itemsInSubPlot.index(plotId)

    def plotIdsInSubPlot(self, subPlotId: int) -> List[int]:
        """return all plot IDs in a given subplot

        :param subPlotId: ID of the subplot
        :return: list of plot IDs
        """
        itemsInSubPlot = [i for i in self.allPlotIds if self.plotItems[i].subPlot == subPlotId]
        return itemsInSubPlot

    def dataDimensionsInSubPlot(self, subPlotId: int) -> Dict[int, int]:
        """Determine what the data dimensions are in a subplot.

        :param subPlotId: ID of the subplot
        :return: dictionary with plot id as key, data dimension (i.e., number of independents) as value.
        """
        ret: Dict[int, int] = {}
        for plotId in self.plotIdsInSubPlot(subPlotId):
            ret[plotId] = len(self.plotItems[plotId].data) - 1
        return ret

    # Methods to be implemented by inheriting classes
    def makeSubPlots(self, nSubPlots: int) -> List[Any]:
        """Generate the subplots. Called after all data has been added.
        Must be implemented by an inheriting class.

        :param nSubPlots: number of subplots
        :return: return values of the subplot generation methods.
        """
        raise NotImplementedError

    def formatSubPlot(self, subPlotId: int) -> Any:
        """Format a subplot.
        May be implemented by an inheriting class.
        By default, does nothing.

        :param subPlotId: ID of the subplot.
        :return: Depends on inheriting class.
        """
        return None

    def plot(self, plotItem: PlotItem) -> Any:
        """Plot an item.
        Must be implemented by an inheriting class.

        :param plotItem: the item to plot.
        :return: Depends on the inheriting class.
        """
        raise NotImplementedError


def _generate_auto_dict_key(d: Dict) -> int:
    guess = 0
    while guess in d.keys():
        guess += 1
    return guess



"""
Generation of useful test data.

Convention:
functions with name `generate_[...]` return generators that can be
used to produce data line by line.
functions with name `get_[...]` return valid DataDict objects.
"""
from typing import Iterable, Dict
import numpy as np

from plottr.data.datadict import DataDict


def generate_2d_scalar_simple(nx: int, ny: int, ndeps: int = 1) -> Iterable[Dict[str, float]]:
    """
    Generate 2d example data, with axes x, y.

    Axes:
    x = np.arange(0, nx)
    y = np.arange(0, ny)

    Dependents:
    z_<idx> = x * y * <idx>

    :param nx: number of points on the x-axes.
    :param ny: number of points on the y-axis.
    :param ndeps: number of columns
    :param noise: how much noise to add to data
    """

    x = np.arange(nx, dtype=float)
    y = np.arange(ny, dtype=float)
    xx, yy = np.meshgrid(x, y, indexing='ij')

    for x, y in zip(xx.reshape(-1), yy.reshape(-1)):
        ret = dict(x=float(x), y=float(y))
        for n in range(ndeps):
            ret[f'z_{n}'] = float(x * y * n)
        yield ret


def get_1d_scalar_cos_data(nx: int = 10, ndata: int = 1) -> DataDict:
    """
    return a datadict with `ndata` dependents.
    All have a cos-dependence on x (with increasing frequency).
    Also noise is added on top.
    """
    x = np.linspace(0, 10, nx)
    d = DataDict(
        x=dict(values=x, unit='A')
    )
    for n in range(ndata):
        dd = np.cos((n+1)*x) + (-0.1 + 0.2 * np.random.rand(x.size))
        d[f"data_{n+1}"] = dict(values=dd, axes=['x',], unit='a.u.')

    d.validate()
    return d


def get_2d_scalar_cos_data(nx: int = 10, ny: int = 10, ndata: int = 1) -> DataDict:
    """
    return a datadict with `ndata` dependents.
    All have a cos-dependence on x (with increasing frequency),
    and repetitions along y.
    Also noise is added on top.
    """
    x = np.linspace(0, 10, nx)
    y = np.arange(ny)
    xx, yy = np.meshgrid(x, y, indexing='ij')

    d = DataDict(
        x=dict(values=xx.reshape(-1), unit='A'),
        y=dict(values=yy.reshape(-1), unit='B'),
    )
    for n in range(ndata):
        dd = np.cos((n+1)*xx) + (-0.1 + 0.2 * np.random.rand(*yy.shape))
        d[f"data_{n+1}"] = dict(values=dd.reshape(-1), axes=['x', 'y'])

    d.validate()
    return d


# DEPRECATED BUT STILL IN USE
def two_1d_traces(nvals: int = 11) -> DataDict:
    x = np.linspace(0, 10, nvals)
    y = np.cos(x)
    z = np.cos(x) ** 2
    d = DataDict(
        x={'values': x},
        y={'values': y, 'axes': ['x']},
        z={'values': z, 'axes': ['x']},
    )
    d.validate()
    return d


def one_2d_set(nx: int = 10, ny: int = 10) -> DataDict:
    x = np.linspace(0, 10, nx)
    y = np.arange(ny)

    xx, yy = np.meshgrid(x, y, indexing='ij')
    dd = np.cos(xx) + (-0.05 + 0.1 * np.random.rand(*yy.shape))

    d = DataDict(
        x=dict(values=xx.reshape(-1)),
        y=dict(values=yy.reshape(-1)),
        cos_data=dict(values=dd.reshape(-1), axes=['x', 'y']),
    )
    d.validate()
    return d


def two_compatible_noisy_2d_sets(nx: int =10, ny: int = 10) -> DataDict:
    x = np.linspace(0, 10, nx)
    y = np.arange(ny)

    xx, yy = np.meshgrid(x, y, indexing='ij')
    dd = np.cos(xx) + (-0.05 + 0.1 * np.random.rand(*yy.shape))
    dd2 = np.sin(xx) + (-0.5 + 1 * np.random.rand(*yy.shape))

    d = DataDict(
        x=dict(values=xx.reshape(-1)),
        y=dict(values=yy.reshape(-1)),
        cos_data=dict(values=dd.reshape(-1), axes=['x', 'y']),
        sin_data=dict(values=dd2.reshape(-1), axes=['x', 'y']),
    )
    d.validate()
    return d


def three_compatible_3d_sets(nx: int = 3, ny: int = 3,
                             nz: int = 3, rand_factor: int = 1) -> DataDict:
    x = np.linspace(0, 10, nx)
    y = np.linspace(-5, 5, ny)
    z = np.arange(nz)
    xx, yy, zz = np.meshgrid(x, y, z, indexing='ij')
    dd = np.cos(xx) * np.sin(yy) + rand_factor * np.random.rand(*zz.shape)
    dd2 = np.sin(xx) * np.cos(yy) + rand_factor * np.random.rand(*zz.shape)
    dd3 = np.cos(xx) ** 2 * np.cos(yy) ** 2 + rand_factor * np.random.rand(
        *zz.shape)

    d = DataDict(
        x=dict(values=xx.reshape(-1), unit='mA'),
        y=dict(values=yy.reshape(-1), unit='uC'),
        z=dict(values=zz.reshape(-1), unit='nF'),
        data=dict(values=dd.reshape(-1), axes=['x', 'y', 'z'], unit='kW'),
        more_data=dict(values=dd2.reshape(-1), axes=['x', 'y', 'z'], unit='MV'),
        different_data=dict(values=dd3.reshape(-1), axes=['x', 'y', 'z'],
                            unit='TS')
    )
    d.validate()
    return d


def three_incompatible_3d_sets(nx: int = 3, ny: int = 3,
                               nz: int = 3, rand_factor: int = 1) -> DataDict:
    x = np.linspace(0, 10, nx)
    y = np.linspace(-5, 5, ny)
    z = np.arange(nz)
    xx, yy, zz = np.meshgrid(x, y, z, indexing='ij')
    dd = np.cos(xx) * np.sin(yy) + rand_factor * np.random.rand(*zz.shape)
    dd2 = np.sin(xx) * np.cos(yy) + rand_factor * np.random.rand(*zz.shape)
    dd3 = np.cos(xx) ** 2 * np.exp(-yy**2 * 0.2) + rand_factor * np.random.rand(
        *zz.shape)

    d = DataDict(
        x=dict(values=xx.reshape(-1), unit='mA'),
        y=dict(values=yy.reshape(-1), unit='uC'),
        z=dict(values=zz.reshape(-1), unit='nF'),
        data=dict(values=dd.reshape(-1),
                  axes=['x', 'y', 'z'], unit='kW'),
        more_data=dict(values=dd2.reshape(-1),
                       axes=['y', 'x', 'z'], unit='MV'),
        different_data=dict(values=dd3.reshape(-1),
                            axes=['z', 'y', 'x'], unit='TS'),
    )
    d.validate()
    return d



"""misc.py

Various utility functions.
"""

from enum import Enum
from typing import List, Tuple, TypeVar, Optional, Sequence, Any


def reorder_indices(lst: Sequence[str], target: Sequence[str]) -> Tuple[int, ...]:
    """
    Determine how to bring a list with unique entries to a different order.

    Supports only lists of strings.

    :param lst: input list
    :param target: list in the desired order
    :return: the indices that will reorder the input to obtain the target.
    :raises: ``ValueError`` for invalid inputs.
    """
    if set([type(i) for i in lst]) != {str}:
        raise ValueError('Only lists of strings are supported')
    if len(set(lst)) < len(lst):
        raise ValueError('Input list elements are not unique.')
    if set(lst) != set(target) or len(lst) != len(target):
        raise ValueError('Contents of input and target do not match.')

    idxs = []
    for elt in target:
        idxs.append(lst.index(elt))

    return tuple(idxs)


def reorder_indices_from_new_positions(lst: List[str], **pos: int) \
        -> Tuple[int, ...]:
    """
    Determine how to bring a list with unique entries to a different order.

    :param lst: input list (of strings)
    :param pos: new positions in the format ``element = new_position``.
                non-specified elements will be adjusted automatically.
    :return: the indices that will reorder the input to obtain the target.
    :raises: ``ValueError`` for invalid inputs.
    """
    if set([type(i) for i in lst]) != {str}:
        raise ValueError('Only lists of strings are supported')
    if len(set(lst)) < len(lst):
        raise ValueError('Input list elements are not unique.')

    target = lst.copy()
    for item, newidx in pos.items():
        oldidx = target.index(item)
        del target[oldidx]
        target.insert(newidx, item)

    return reorder_indices(lst, target)


T = TypeVar('T')


def unwrap_optional(val: Optional[T]) -> T:
    """Covert a variable of type Optional[T] to T
    If the variable has value None a ValueError will be raised
    """
    if val is None:
        raise ValueError("Expected a not None value but got a None value.")
    return val


class AutoEnum(Enum):
    """Enum that with automatically incremented integer values.

    Allows to pass additional arguments in the class variables to the __init__
    method of the instances.
    See: https://stackoverflow.com/questions/19330460/how-do-i-put-docstrings-on-enums/19330461#19330461
    """

    def __new__(cls, *args: Any) -> "AutoEnum":
        """creating a new instance.

        :param args: will be passed to __init__.
        """
        value = len(cls) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj


class LabeledOptions(AutoEnum):
    """Enum with a label for each element. We can find the name from the label
    using :meth:`.fromLabel`.

    Example::

            >>> class Color(LabeledOptions):
            ...     red = 'Red'
            ...     blue = 'Blue'

    Here, ``Color.blue`` has value ``2`` and ``Color.fromLabel('Blue')`` returns
    ``Color.blue``.
    """

    def __init__(self, label: str) -> None:
        self.label = label

    @classmethod
    def fromLabel(cls, label: str) -> Optional["LabeledOptions"]:
        """Find enum element from label."""
        for k in cls:
            if k.label.lower() == label.lower():
                return k
        return None
